"""
Data analysis code for compressive gate set tomography
"""

from itertools import product
from time import perf_counter
from typing import Any, List, Tuple, Union

from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from matplotlib.transforms import Bbox
from numpy import ndarray
import numpy as np
from pandas import DataFrame
from pygsti.models.model import Model
from pygsti.tools import change_basis
import xarray as xr

from iqm.benchmarks.benchmark_definition import (
    BenchmarkAnalysisResult,
    BenchmarkObservation,
    BenchmarkObservationIdentifier,
    BenchmarkRunResult,
)
from mGST import additional_fns, algorithm, compatibility
from mGST.low_level_jit import contract
from mGST.qiskit_interface import qiskit_gate_to_operator
from mGST.reporting import figure_gen, reporting


def dataframe_to_figure(
    df: DataFrame, row_labels: Union[List[str], None] = None, col_width: float = 2, fontsize: int = 12
) -> Figure:
    """Turns a pandas DataFrame into a figure
    This is needed to conform with the standard file saving routine of QCVV.

    Args:
        df: Pandas DataFrame
            A dataframe table containing GST results
        row_labels: List[str]
            The row labels for the dataframe
        col_width: int
            Used to control cell width in the table
        fontsize: int
            Font size of text/numbers in table cells

    Returns:
        figure: Matplotlib figure object
            A figure representing the dataframe.
    """

    if row_labels is None:
        row_labels = list(np.arange(df.shape[0]))

    row_height = fontsize / 70 * 2
    n_cols = df.shape[1]
    n_rows = df.shape[0]
    figsize = np.array([n_cols + 1, n_rows + 1]) * np.array([col_width, row_height])

    fig, ax = plt.subplots(figsize=figsize)

    fig.patch.set_visible(False)
    ax.axis("off")
    ax.axis("tight")
    data_array = (df.to_numpy(dtype="str")).copy()
    column_names = df.columns.tolist()
    table = ax.table(
        cellText=data_array,
        colLabels=column_names,
        rowLabels=row_labels,
        cellLoc="center",
        colColours=["#7FA1C3" for _ in range(n_cols)],
        bbox=Bbox([[0, 0], [1, 1]]),
    )
    table.set_fontsize(fontsize)
    table.set_figure(fig)
    return fig


def bootstrap_errors(
    dataset: xr.Dataset,
    y: ndarray,
    K: ndarray,
    X: ndarray,
    E: ndarray,
    rho: ndarray,
    target_mdl: Model,
    parametric: bool = False,
) -> tuple[Any, Any, Any, Any, Any]:
    """Resamples circuit outcomes a number of times and computes GST estimates for each repetition
    All results are then returned in order to compute bootstrap-error bars for GST estimates.
    Parametric bootstrapping uses the estimated gate set to create a newly sampled data set.
    Non-parametric bootstrapping uses the initial dataset and resamples according to the
    corresp. outcome probabilities.
    Each bootstrap run is initialized with the estimated gate set in order to save processing time.

    Parameters
    ----------
    dataset: xarray.Dataset
        A dataset containing counts from the experiment and configurations
    qubit_layout: List[int]
        The list of qubits for the current GST experiment
    y: ndarray
        The circuit outcome probabilities as a num_povm x num_circuits array
    K : ndarray
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.
    X : 3D ndarray
        Array where reconstructed CPT superoperators in standard basis are stacked along the first axis.
    E : ndarray
        Current POVM estimate
    rho : ndarray
        Current initial state estimate
    target_mdl : pygsti model object
        The target gate set
    parametric : bool
        If set to True, parametric bootstrapping is used, else non-parametric bootstrapping. Default: False

    Returns
    -------
    X_array : ndarray
        Array containing all estimated gate tensors of different bootstrapping repetitions along first axis
    E_array : ndarray
        Array containing all estimated POVM tensors of different bootstrapping repetitions along first axis
    rho_array : ndarray
        Array containing all estimated initial states of different bootstrapping repetitions along first axis
    df_g_array : ndarray
        Contains gate quality measures of bootstrapping repetitions
    df_o_array : ndarray
        Contains SPAM and other quality measures of bootstrapping repetitions

    """
    if parametric:
        y = np.real(
            np.array(
                [
                    [E[i].conj() @ contract(X, j) @ rho for j in dataset.attrs["J"]]
                    for i in range(dataset.attrs["num_povm"])
                ]
            )
        )
    X_array = np.zeros((dataset.attrs["bootstrap_samples"], *X.shape)).astype(complex)
    E_array = np.zeros((dataset.attrs["bootstrap_samples"], *E.shape)).astype(complex)
    rho_array = np.zeros((dataset.attrs["bootstrap_samples"], *rho.shape)).astype(complex)
    df_g_list = []
    df_o_list = []

    for i in range(dataset.attrs["bootstrap_samples"]):
        y_sampled = additional_fns.sampled_measurements(y, dataset.attrs["shots"]).copy()
        _, X_, E_, rho_, _ = algorithm.run_mGST(
            y_sampled,
            dataset.attrs["J"],
            dataset.attrs["seq_len_list"][-1],
            dataset.attrs["num_gates"],
            dataset.attrs["pdim"] ** 2,
            dataset.attrs["rank"],
            dataset.attrs["num_povm"],
            dataset.attrs["batch_size"],
            dataset.attrs["shots"],
            method=dataset.attrs["opt_method"],
            max_inits=dataset.attrs["max_inits"],
            max_iter=0,
            final_iter=dataset.attrs["max_iterations"][1],
            threshold_multiplier=dataset.attrs["convergence_criteria"][0],
            target_rel_prec=dataset.attrs["convergence_criteria"][1],
            init=[K, E, rho],
            testing=False,
        )

        X_opt, E_opt, rho_opt = reporting.gauge_opt(X_, E_, rho_, target_mdl, dataset.attrs[f"gauge_weights"])
        df_g, df_o = reporting.report(
            X_opt,
            E_opt,
            rho_opt,
            dataset.attrs["J"],
            y_sampled,
            target_mdl,
            dataset.attrs["gate_labels"],
        )
        df_g_list.append(df_g.values)
        df_o_list.append(df_o.values)

        X_opt_pp, E_opt_pp, rho_opt_pp = compatibility.std2pp(X_opt, E_opt, rho_opt)

        X_array[i] = X_opt_pp
        E_array[i] = E_opt_pp
        rho_array[i] = rho_opt_pp

    return X_array, E_array, rho_array, np.array(df_g_list), np.array(df_o_list)


def generate_non_gate_results(
    dataset: xr.Dataset, qubit_layout: List[int], df_o: DataFrame
) -> tuple[DataFrame, Figure]:
    """
    Creates error bars (if bootstrapping was used) and formats results for non-gate errors.
    The resulting tables are also turned into figures, so that they can be saved automatically.

    Args:
        dataset: xr.Dataset
            A dataset containing counts from the experiment and configurations
        qubit_layout: List[int]
                The list of qubits for the current GST experiment
        df_o: Pandas DataFrame
            A dataframe containing the non-gate quality metrics (SPAM errors and fit quality)

    Returns:
        df_o_final: Pandas DataFrame
            The final formated results
    """
    identifier = BenchmarkObservationIdentifier(qubit_layout).string_identifier
    if dataset.attrs["bootstrap_samples"] > 0:
        _, _, _, _, df_o_array = dataset.attrs["results_layout_" + identifier]["bootstrap_data"]
        df_o_array[df_o_array == -1] = np.nan
        percentiles_o_low, percentiles_o_high = np.nanpercentile(df_o_array, [2.5, 97.5], axis=0)
        df_o_final = DataFrame(
            {
                f"Mean TVD: estimate - data": reporting.number_to_str(
                    df_o.values[0, 1].copy(), [percentiles_o_high[0, 1], percentiles_o_low[0, 1]], precision=5
                ),
                f"Mean TVD: target - data": reporting.number_to_str(
                    df_o.values[0, 2].copy(), [percentiles_o_high[0, 2], percentiles_o_low[0, 2]], precision=5
                ),
                f"POVM - diamond dist.": reporting.number_to_str(
                    df_o.values[0, 3].copy(), [percentiles_o_high[0, 3], percentiles_o_low[0, 3]], precision=5
                ),
                f"State - trace dist.": reporting.number_to_str(
                    df_o.values[0, 4].copy(), [percentiles_o_high[0, 4], percentiles_o_low[0, 4]], precision=5
                ),
            },
            index=[""],
        )
    else:
        df_o_final = DataFrame(
            {
                f"Mean TVD: estimate - data": reporting.number_to_str(df_o.values[0, 1].copy(), precision=5),
                f"Mean TVD: target - data": reporting.number_to_str(df_o.values[0, 2].copy(), precision=5),
                f"POVM - diamond dist.": reporting.number_to_str(df_o.values[0, 3].copy(), precision=5),
                f"State - trace dist.": reporting.number_to_str(df_o.values[0, 4].copy(), precision=5),
            },
            index=[""],
        )
    fig = dataframe_to_figure(df_o_final, [""])  # dataframe_to_figure(df_o_final, [""])
    return df_o_final, fig


def generate_unit_rank_gate_results(
    dataset: xr.Dataset, qubit_layout: List[int], df_g: DataFrame, X_opt: ndarray, K_target: ndarray
) -> Tuple[DataFrame, DataFrame, Figure, Figure]:
    """
    Produces all result tables for Kraus rank 1 estimates and turns them into figures.

    This includes parameters of the Hamiltonian generators in the Pauli basis for all gates,
    as well as the usual performance metrics (Fidelities and Diamond distances). If bootstrapping
    data is available, error bars will also be generated.

    Args:
        dataset: xarray.Dataset
            A dataset containing counts from the experiment and configurations
        qubit_layout: List[int]
            The list of qubits for the current GST experiment
        df_g: Pandas DataFrame
            The dataframe with properly formated results
        X_opt: 3D numpy array
            The gate set after gauge optimization
        K_target: 4D numpy array
            The Kraus operators of all target gates, used to compute distance measures.

    Returns:
        df_g_final: Pandas DataFrame
            The dataframe with properly formated results of standard gate errors
        df_g_rotation Pandas DataFrame
            A dataframe containing Hamiltonian (rotation) parameters
        fig_g: Figure
            A table in Figure format of gate results (fidelities etc.)
        fig_rotation: Figure
            A table in Figure format of gate Hamiltonian parameters


    """
    identifier = BenchmarkObservationIdentifier(qubit_layout).string_identifier
    pauli_labels = generate_basis_labels(dataset.attrs["pdim"], basis="Pauli")
    if dataset.attrs["bootstrap_samples"] > 0:
        X_array, E_array, rho_array, df_g_array, _ = dataset.attrs["results_layout_" + identifier]["bootstrap_data"]
        df_g_array[df_g_array == -1] = np.nan
        percentiles_g_low, percentiles_g_high = np.nanpercentile(df_g_array, [2.5, 97.5], axis=0)

        df_g_final = DataFrame(
            {
                r"Avg. gate fidelity": [
                    reporting.number_to_str(
                        df_g.values[i, 0], [percentiles_g_high[i, 0], percentiles_g_low[i, 0]], precision=5
                    )
                    for i in range(len(dataset.attrs["gate_labels"]))
                ],
                r"Diamond distance": [
                    reporting.number_to_str(
                        df_g.values[i, 1], [percentiles_g_high[i, 1], percentiles_g_low[i, 1]], precision=5
                    )
                    for i in range(dataset.attrs["num_gates"])
                ],
            }
        )

        U_opt = reporting.phase_opt(X_opt, K_target)
        pauli_coeffs = reporting.compute_sparsest_Pauli_Hamiltonian(U_opt)

        bootstrap_pauli_coeffs = np.zeros((len(X_array), dataset.attrs["num_gates"], dataset.attrs["pdim"] ** 2))
        for i, X_ in enumerate(X_array):
            X_std, _, _ = compatibility.pp2std(X_, E_array[i], rho_array[i])
            U_opt_ = reporting.phase_opt(
                np.array([change_basis(X_std[j], "pp", "std") for j in range(dataset.attrs["num_gates"])]), K_target
            )
            pauli_coeffs_ = reporting.compute_sparsest_Pauli_Hamiltonian(U_opt_)
            bootstrap_pauli_coeffs[i, :, :] = pauli_coeffs_
        pauli_coeffs_low, pauli_coeffs_high = np.nanpercentile(bootstrap_pauli_coeffs, [2.5, 97.5], axis=0)

        df_g_rotation = DataFrame(
            np.array(
                [
                    [
                        reporting.number_to_str(
                            pauli_coeffs[i, j], [pauli_coeffs_high[i, j], pauli_coeffs_low[i, j]], precision=5
                        )
                        for i in range(dataset.attrs["num_gates"])
                    ]
                    for j in range(dataset.attrs["pdim"] ** 2)
                ]
            ).T
        )

        df_g_rotation.columns = [f"h_%s" % label for label in pauli_labels]
        df_g_rotation.rename(index=dataset.attrs["gate_labels"], inplace=True)

    else:
        df_g_final = DataFrame(
            {
                "Avg. gate fidelity": [
                    reporting.number_to_str(df_g.values[i, 0], precision=5) for i in range(dataset.attrs["num_gates"])
                ],
                "Diamond distance": [
                    reporting.number_to_str(df_g.values[i, 1], precision=5) for i in range(dataset.attrs["num_gates"])
                ],
            }
        )
        U_opt = reporting.phase_opt(X_opt, K_target)
        pauli_coeffs = reporting.compute_sparsest_Pauli_Hamiltonian(U_opt)

        df_g_rotation = DataFrame(
            np.array(
                [
                    [
                        reporting.number_to_str(pauli_coeffs[i, j], precision=5)
                        for i in range(dataset.attrs["num_gates"])
                    ]
                    for j in range(dataset.attrs["pdim"] ** 2)
                ]
            ).T
        )
        df_g_rotation.columns = [f"h_%s" % label for label in pauli_labels]
        df_g_final.rename(index=dataset.attrs["gate_labels"], inplace=True)

    fig_g = dataframe_to_figure(df_g_final, dataset.attrs["gate_labels"])
    fig_rotation = dataframe_to_figure(df_g_rotation, dataset.attrs["gate_labels"])
    return df_g_final, df_g_rotation, fig_g, fig_rotation


def generate_gate_results(
    dataset: xr.Dataset,
    qubit_layout: List[int],
    df_g: DataFrame,
    X_opt: ndarray,
    E_opt: ndarray,
    rho_opt: ndarray,
    max_evals: int = 6,
) -> Tuple[DataFrame, DataFrame, Figure, Figure]:
    """
    Produces all result tables for arbitrary Kraus rank estimates and turns them into figures.

    Args:
        df_g: Pandas DataFrame
            The dataframe with properly formated results
        X_opt: 3D numpy array
            The gate set after gauge optimization
        E_opt: 3D numpy array
            An array containg all the POVM elements as matrices after gauge optimization
        rho_opt: 2D numpy array
            The density matrix after gauge optmization
        max_evals: int
            The maximum number of eigenvalues of the Choi matrices which are returned.

    Returns:
        df_g_final: Pandas DataFrame
            The dataframe with properly formated results of standard gate errors
        df_g_evals_final Pandas DataFrame
            A dataframe containing eigenvalues of the Choi matrices for all gates
        fig_g: Figure
            A table in Figure format of gate results (fidelities etc.)
        fig_choi: Figure
            A table in Figure format of eigenvalues of the Choi matrices of all gates

    """
    identifier = BenchmarkObservationIdentifier(qubit_layout).string_identifier
    n_evals = np.min([max_evals, dataset.attrs["pdim"] ** 2])
    X_opt_pp, _, _ = compatibility.std2pp(X_opt, E_opt, rho_opt)
    df_g_evals = reporting.generate_Choi_EV_table(X_opt, n_evals, dataset.attrs["gate_labels"])

    if dataset.attrs["bootstrap_samples"] > 0:
        X_array, E_array, rho_array, df_g_array, _ = dataset.attrs["results_layout_" + identifier]["bootstrap_data"]
        df_g_array[df_g_array == -1] = np.nan
        percentiles_g_low, percentiles_g_high = np.nanpercentile(df_g_array, [2.5, 97.5], axis=0)
        bootstrap_unitarities = np.array(
            [reporting.unitarities(X_array[i]) for i in range(dataset.attrs["bootstrap_samples"])]
        )
        percentiles_u_low, percentiles_u_high = np.nanpercentile(bootstrap_unitarities, [2.5, 97.5], axis=0)
        X_array_std = [
            compatibility.pp2std(X_array[i], E_array[i], rho_array[i])[0]
            for i in range(dataset.attrs["bootstrap_samples"])
        ]
        bootstrap_evals = np.array(
            [
                reporting.generate_Choi_EV_table(X_array_std[i], n_evals, dataset.attrs["gate_labels"])
                for i in range(dataset.attrs["bootstrap_samples"])
            ]
        )
        percentiles_evals_low, percentiles_evals_high = np.nanpercentile(bootstrap_evals, [2.5, 97.5], axis=0)
        eval_strs = [
            [
                reporting.number_to_str(
                    df_g_evals.values[i, j],
                    [percentiles_evals_high[i, j], percentiles_evals_low[i, j]],
                    precision=5,
                )
                for i in range(dataset.attrs["num_gates"])
            ]
            for j in range(n_evals)
        ]

        df_g_final = DataFrame(
            {
                r"Avg. gate fidelity": [
                    reporting.number_to_str(
                        df_g.values[i, 0], [percentiles_g_high[i, 0], percentiles_g_low[i, 0]], precision=5
                    )
                    for i in range(dataset.attrs["num_gates"])
                ],
                r"Diamond distance": [
                    reporting.number_to_str(
                        df_g.values[i, 1], [percentiles_g_high[i, 1], percentiles_g_low[i, 1]], precision=5
                    )
                    for i in range(dataset.attrs["num_gates"])
                ],
                r"Unitarity": [
                    reporting.number_to_str(
                        reporting.unitarities(X_opt_pp)[i],
                        [percentiles_u_high[i], percentiles_u_low[i]],
                        precision=5,
                    )
                    for i in range(dataset.attrs["num_gates"])
                ],
            }
        )

    else:
        df_g_final = DataFrame(
            {
                "Avg. gate fidelity": [
                    reporting.number_to_str(df_g.values[i, 0].copy(), precision=5)
                    for i in range(len(dataset.attrs["gate_labels"]))
                ],
                "Diamond distance": [
                    reporting.number_to_str(df_g.values[i, 1].copy(), precision=5)
                    for i in range(len(dataset.attrs["gate_labels"]))
                ],
                "Unitarity": [
                    reporting.number_to_str(reporting.unitarities(X_opt_pp)[i], precision=5)
                    for i in range(len(dataset.attrs["gate_labels"]))
                ],
                # "Entanglemen fidelity to depol. channel": [reporting.number_to_str(reporting.eff_depol_params(X_opt_pp)[i], precision=5)
                #                                            for i in range(len(gate_labels))],
                # "Min. spectral distances": [number_to_str(df_g.values[i, 2], precision=5) for i in range(len(gate_labels))]
            }
        )
        eval_strs = [
            [
                reporting.number_to_str(df_g_evals.values[i, j].copy(), precision=5)
                for i in range(dataset.attrs["num_gates"])
            ]
            for j in range(n_evals)
        ]

    df_g_evals_final = DataFrame(eval_strs).T
    df_g_evals_final.rename(index=dataset.attrs["gate_labels"], inplace=True)

    fig_g = dataframe_to_figure(df_g_final, dataset.attrs["gate_labels"])
    fig_choi = dataframe_to_figure(df_g_evals_final, dataset.attrs["gate_labels"])
    return df_g_final, df_g_evals_final, fig_g, fig_choi


def generate_basis_labels(pdim: int, basis: Union[str, None] = None) -> List[str]:
    """Generate a list of labels for the Pauli basis or the standard basis

    Args:
        pdim: int
            Physical dimension
        basis: str
            Which basis the labels correspond to, currently default is standard basis and "Pauli" can be choose
            for Pauli basis labels like "II", "IX", "XX", ...

    Returns:
        labels: List[str]
            A list of all string combinations for the given dimension and basis
    """
    separator = ""
    if basis == "Pauli":
        pauli_labels_loc = ["I", "X", "Y", "Z"]
        pauli_labels_rep = [pauli_labels_loc for _ in range(int(np.log2(pdim)))]
        labels = [separator.join(map(str, x)) for x in product(*pauli_labels_rep)]
    else:
        std_labels_loc = ["0", "1"]
        std_labels_rep = [std_labels_loc for _ in range(int(np.log2(pdim)))]
        labels = [separator.join(map(str, x)) for x in product(*std_labels_rep)]

    return labels


def result_str_to_floats(result_str: str, err: str) -> Tuple[float, float]:
    """Converts formated string results from mgst to float (value, uncertainty) pairs

    Args:
        result_str: str
            The value of a result parameter formated as str
        err: str
            The error interval of the parameters

    Returns:
        value: float
            The parameter value as floar
        uncertainty: float
            A single uncertainty value
    """
    if err:
        value = float(result_str.split("[")[0])
        rest = result_str.split("[")[1].split(",")
        uncertainty = float(rest[1][:-1]) - float(rest[0])
        return value, uncertainty
    return float(result_str), np.NaN


def pandas_results_to_observations(
    dataset: xr.Dataset, df_g: DataFrame, df_o: DataFrame, identifier: BenchmarkObservationIdentifier
) -> List[BenchmarkObservation]:
    """Converts high level GST results from a pandas Dataframe to a simple observation dictionary

    Args:
    dataset: xarray.Dataset
        A dataset containing counts from the experiment and configurations
    qubit_layout: List[int]
        The list of qubits for the current GST experiment
    df_g: Pandas DataFrame
        The dataframe with properly formated gate results
    df_o: Pandas DataFrame
        The dataframe with properly formated non-gate results like SPAM error measures or fit quality.
    identifier: BenchmarkObservationIdentifier
        An identifier object for the current GST run

    Returns:
    observation_list: List[BenchmarkObservation]
        List of observations converted from the pandas dataframes
    """
    observation_list: list[BenchmarkObservation] = []
    err = dataset.attrs["bootstrap_samples"] > 0
    for idx, gate_label in enumerate(dataset.attrs["gate_labels"].values()):
        observation_list.extend(
            [
                BenchmarkObservation(
                    name=f"{name} - {gate_label}",
                    identifier=identifier,
                    value=result_str_to_floats(df_g[name].iloc[idx], err)[0],
                    uncertainty=result_str_to_floats(df_g[name].iloc[idx], err)[1],
                )
                for name in df_g.columns.tolist()
            ]
        )
    observation_list.extend(
        [
            BenchmarkObservation(
                name=f"{name}",
                identifier=identifier,
                value=result_str_to_floats(df_o[name].iloc[0], err)[0],
                uncertainty=result_str_to_floats(df_o[name].iloc[0], err)[1],
            )
            for name in df_o.columns.tolist()
        ]
    )
    return observation_list


def dataset_counts_to_mgst_format(dataset: xr.Dataset, qubit_layout: List[int]) -> ndarray:
    """Turns the dictionary of outcomes obtained from qiskit backend
        into the format which is used in mGST

    Args:
        dataset: xarray.Dataset
            A dataset containing counts from the experiment and configurations
        qubit_layout: List[int]
            The list of qubits for the current GST experiment

    Returns
    -------
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence

    """
    num_qubits = len(qubit_layout)
    num_povm = dataset.attrs["num_povm"]
    y_list = []
    for run_index in range(dataset.attrs["num_circuits"]):
        # result = dataset[f"{qubit_layout}_counts_{run_index}"].to_dict()
        result = dataset[f"{qubit_layout}_state_{run_index}"].data.tolist()
        basis_dict = {entry: int("".join([entry[::-1][i] for i in range(num_qubits)][::-1]), 2) for entry in result}
        # Sort by index:
        basis_dict = dict(sorted(basis_dict.items(), key=lambda item: item[1]))

        counts_normalized = (
            dataset[f"{qubit_layout}_counts_{run_index}"] / dataset[f"{qubit_layout}_counts_{run_index}"].sum()
        )
        row = [counts_normalized.loc[key].data for key in basis_dict]
        # row = [result[key] for key in basis_dict]
        if len(row) < num_povm:
            missing_entries = list(np.arange(num_povm))
            for given_entry in basis_dict.values():
                missing_entries.remove(given_entry)
            for missing_entry in missing_entries:
                row.insert(missing_entry, 0)  # 0 measurement outcomes in not recorded entry
        y_list.append(row)
    y = np.array(y_list).T
    return y


def run_mGST(
    dataset: xr.Dataset, y: ndarray
) -> tuple[ndarray, ndarray, ndarray, ndarray, ndarray, ndarray, ndarray, ndarray]:
    """Wrapper function for mGST algorithm execution which prepares an initialization and sets the alg. parameters

    Args:
        dataset: xarray.Dataset
            A dataset containing counts from the experiment and configurations
        y: ndarray
            The circuit outcome probabilities as a num_povm x num_circuits array

    Returns:
        K : ndarray
            Kraus estimate array where each subarray along the first axis contains a set of Kraus operators.
            The second axis enumerates Kraus operators for a gate specified by the first axis.
        X : ndarray
            Superoperator estimate array where reconstructed CPT superoperators in
            standard basis are stacked along the first axis.
        E : ndarray
            Current POVM estimate
        rho : ndarray
            Current initial state estimate
        K_target : ndarray
            Target gate Kraus array where each subarray along the first axis contains a set of Kraus operators.
            The second axis enumerates Kraus operators for a gate specified by the first axis.
        X_target : ndarray
            Target gate superoperator estimate array where reconstructed CPT superoperators in
            standard basis are stacked along the first axis.
        E_target : ndarray
            Target POVM
        rho_target : ndarray
            Target initial state
    """

    K_target = qiskit_gate_to_operator(dataset.attrs["gate_set"])
    X_target = np.einsum("ijkl,ijnm -> iknlm", K_target, K_target.conj()).reshape(
        (dataset.attrs["num_gates"], dataset.attrs["pdim"] ** 2, dataset.attrs["pdim"] ** 2)
    )  # tensor of superoperators

    rho_target = (
        np.kron(additional_fns.basis(dataset.attrs["pdim"], 0).T.conj(), additional_fns.basis(dataset.attrs["pdim"], 0))
        .reshape(-1)
        .astype(np.complex128)
    )

    # Computational basis measurement:
    E_target = np.array(
        [
            np.kron(
                additional_fns.basis(dataset.attrs["pdim"], i).T.conj(), additional_fns.basis(dataset.attrs["pdim"], i)
            ).reshape(-1)
            for i in range(dataset.attrs["pdim"])
        ]
    ).astype(np.complex128)

    # Run mGST
    if dataset.attrs["from_init"]:
        K_init = additional_fns.perturbed_target_init(X_target, dataset.attrs["rank"])
        init_params = [K_init, E_target, rho_target]
    else:
        init_params = None

    K, X, E, rho, _ = algorithm.run_mGST(
        y,
        dataset.attrs["J"],
        dataset.attrs["seq_len_list"][-1],
        dataset.attrs["num_gates"],
        dataset.attrs["pdim"] ** 2,
        dataset.attrs["rank"],
        dataset.attrs["num_povm"],
        dataset.attrs["batch_size"],
        dataset.attrs["shots"],
        method=dataset.attrs["opt_method"],
        max_inits=dataset.attrs["max_inits"],
        max_iter=dataset.attrs["max_iterations"][0],
        final_iter=dataset.attrs["max_iterations"][1],
        threshold_multiplier=dataset.attrs["convergence_criteria"][0],
        target_rel_prec=dataset.attrs["convergence_criteria"][1],
        init=init_params,
        testing=False,
    )

    return K, X, E, rho, K_target, X_target, E_target, rho_target


def mgst_analysis(run: BenchmarkRunResult) -> BenchmarkAnalysisResult:
    """Analysis function for compressive GST

    Args:
        run: BenchmarkRunResult
            A BenchmarkRunResult instance storing the dataset
    Returns:
        result: BenchmarkAnalysisResult
            An BenchmarkAnalysisResult instance with the updated dataset, as well as plots and observations
    """
    dataset = run.dataset
    pdim = dataset.attrs["pdim"]
    plots = {}
    for i, qubit_layout in enumerate(dataset.attrs["qubit_layouts"]):
        identifier = BenchmarkObservationIdentifier(qubit_layout).string_identifier

        # Computing circuit outcome probabilities from counts
        y = dataset_counts_to_mgst_format(dataset, qubit_layout)

        # Main GST reconstruction
        start_timer = perf_counter()
        K, X, E, rho, K_target, X_target, E_target, rho_target = run_mGST(dataset, y)
        main_gst_time = perf_counter() - start_timer

        # Gauge optimization
        start_timer = perf_counter()
        target_mdl = compatibility.arrays_to_pygsti_model(X_target, E_target, rho_target, basis="std")
        X_opt, E_opt, rho_opt = reporting.gauge_opt(X, E, rho, target_mdl, dataset.attrs[f"gauge_weights"])
        gauge_optimization_time = perf_counter() - start_timer

        # Quick report
        df_g, _ = reporting.quick_report(
            X_opt, E_opt, rho_opt, dataset.attrs["J"], y, target_mdl, dataset.attrs["gate_labels"]
        )

        # Gate set in the Pauli basis
        X_opt_pp, _, _ = compatibility.std2pp(X_opt, E_opt, rho_opt)
        X_target_pp, _, _ = compatibility.std2pp(X_target, E_target, rho_target)

        # Saving
        dataset.attrs["results_layout_" + identifier] = {
            "raw_Kraus_operators": K,
            "raw_gates": X,
            "raw_POVM": E.reshape((dataset.attrs["num_povm"], pdim, pdim)),
            "raw_state": rho.reshape((pdim, pdim)),
            "gauge_opt_gates": X_opt,
            "gauge_opt_gates_Pauli_basis": X_opt_pp,
            "gauge_opt_POVM": E_opt.reshape((dataset.attrs["num_povm"], pdim, pdim)),
            "gauge_opt_state": rho_opt.reshape((pdim, pdim)),
            "main_mGST_time": main_gst_time,
            "gauge_optimization_time": gauge_optimization_time,
        }

        ### Bootstrap
        if dataset.attrs["bootstrap_samples"] > 0:
            bootstrap_results = bootstrap_errors(dataset, y, K, X, E, rho, target_mdl)
            dataset.attrs["results_layout_" + identifier].update({"bootstrap_data": bootstrap_results})

        _, df_o_full = reporting.report(
            X_opt, E_opt, rho_opt, dataset.attrs["J"], y, target_mdl, dataset.attrs["gate_labels"]
        )
        df_o_final, fig_o = generate_non_gate_results(dataset, qubit_layout, df_o_full)

        ### Result table generation and full report
        if dataset.attrs["rank"] == 1:
            df_g_final, df_g_rotation, fig_g, fig_rotation = generate_unit_rank_gate_results(
                dataset, qubit_layout, df_g, X_opt, K_target
            )
            dataset.attrs["results_layout_" + identifier].update({"hamiltonian_parameters": df_g_rotation.to_dict()})
            plots[f"layout_{qubit_layout}_hamiltonian_parameters"] = fig_rotation
        else:
            df_g_final, df_g_evals, fig_g, fig_choi = generate_gate_results(
                dataset, qubit_layout, df_g, X_opt, E_opt, rho_opt
            )
            dataset.attrs["results_layout_" + identifier].update({"choi_evals": df_g_evals.to_dict()})
            plots[f"layout_{qubit_layout}_choi_eigenvalues"] = fig_choi
        plots[f"layout_{qubit_layout}_gate_metrics"] = fig_g
        plots[f"layout_{qubit_layout}_other_metrics"] = fig_o

        observation_list = pandas_results_to_observations(
            dataset, df_g_final, df_o_final, BenchmarkObservationIdentifier(qubit_layout)
        )

        dataset.attrs["results_layout_" + identifier].update(
            {"full_metrics": {"Gates": df_g_final.to_dict(), "Outcomes and SPAM": df_o_final.to_dict()}}
        )

        ### Process matrix plots
        pauli_labels = generate_basis_labels(pdim, basis="Pauli")
        std_labels = generate_basis_labels(pdim)

        figures = figure_gen.generate_gate_err_pdf(
            "",
            X_opt_pp,
            X_target_pp,
            basis_labels=pauli_labels,
            gate_labels=dataset.attrs["gate_labels"],
            return_fig=True,
        )
        for i, figure in enumerate(figures):
            plots[f"layout_{qubit_layout}_process_matrix_{i}"] = figure

        plots[f"layout_{qubit_layout}_SPAM_matrices_real"] = figure_gen.generate_spam_err_std_pdf(
            "",
            E_opt.real,
            rho_opt.real,
            E_target.real,
            rho_target.real,
            basis_labels=std_labels,
            title=f"Real part of state and measurement effects in the standard basis",
            return_fig=True,
        )
        plots[f"layout_{qubit_layout}_SPAM_matrices_imag"] = figure_gen.generate_spam_err_std_pdf(
            "",
            E_opt.imag,
            rho_opt.imag,
            E_target.imag,
            rho_target.imag,
            basis_labels=std_labels,
            title=f"Imaginary part of state and measurement effects in the standard basis",
            return_fig=True,
        )
        plt.close("all")

    return BenchmarkAnalysisResult(dataset=dataset, observations=observation_list, plots=plots)



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Quantum Volume benchmark
"""

from copy import deepcopy
from time import strftime
from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple, Type

from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from mthree.classes import QuasiCollection
from mthree.utils import expval
import numpy as np
from qiskit import QuantumCircuit
from qiskit.circuit.library import QuantumVolume
from qiskit_aer import Aer
import xarray as xr

# import iqm.diqe.executors.dynamical_decoupling.dd_high_level as dd
# from iqm.diqe.executors.dynamical_decoupling.dynamical_decoupling_core import DDStrategy
# from iqm.diqe.mapomatic import evaluate_costs, get_calibration_fidelities, get_circuit, matching_layouts
from iqm.benchmarks.benchmark import BenchmarkConfigurationBase
from iqm.benchmarks.benchmark_definition import (
    Benchmark,
    BenchmarkAnalysisResult,
    BenchmarkObservation,
    BenchmarkObservationIdentifier,
    BenchmarkRunResult,
    add_counts_to_dataset,
)
from iqm.benchmarks.logging_config import qcvv_logger
from iqm.benchmarks.readout_mitigation import apply_readout_error_mitigation
from iqm.benchmarks.utils import (  # execute_with_dd,
    count_native_gates,
    perform_backend_transpilation,
    retrieve_all_counts,
    retrieve_all_job_metadata,
    set_coupling_map,
    sort_batches_by_final_layout,
    submit_execute,
    timeit,
    xrvariable_to_counts,
)
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase


def compute_heavy_output_probabilities(
    execution_results: List[Dict[str, int]],
    ideal_heavy_outputs: List[Dict[str, float]],
) -> List[float]:
    """Compute the HOP of all quantum circuits.

    Args:
        execution_results (List[Dict[str, int]]): counts from execution of all quantum circuits.
        ideal_heavy_outputs (List[Dict[str, float]]): list of ideal heavy output dictionaries.

    Returns:
        List[float]: the HOP of all quantum circuits.
    """
    qv_result = []
    for result, heavy in zip(execution_results, ideal_heavy_outputs):
        qv_result += [expval(result, heavy)]

    return qv_result


def cumulative_hop(hops: List[float]) -> List[float]:
    """
    Computes the cumulative average heavy output probabilities of a QV experiment.
    Args:
        hops (List[float]): The individual heavy output probabilities for each trial.
    Returns:
        List[float]: cumulative average heavy output probabilities for all trials.
    """
    c_h: List[float] = [np.mean(hops[: (i + 1)], dtype=float) for i in range(len(hops))]
    return c_h


def cumulative_std(hops: List[float]) -> List[float]:
    """Computes the cumulative standard deviation heavy output probabilities of a QV experiment.

    Args:
        hops (List[float]): The individual heavy output probabilities for each trial.
    Returns:
        List[float]: cumulative standard deviation heavy output probabilities for all trials.
    """
    c_h = cumulative_hop(hops)
    c_s = [(c_h[i] * ((1 - c_h[i]) / len(hops[: (i + 1)]))) ** 0.5 for i in range(len(hops))]
    return c_s


def get_ideal_heavy_outputs(
    qc_list: List[QuantumCircuit],
    sorted_qc_list_indices: Dict[Tuple[int, ...], List[int]],
) -> List[Dict[str, float]]:
    """Calculate the heavy output bitrstrings of a list of quantum circuits.

    Args:
        qc_list (List[QuantumCircuit]): the list of quantum circuits.
        sorted_qc_list_indices (Dict[Tuple, List[int]]): dictionary of indices (integers) corresponding to those in the original (untranspiled) list of circuits, with keys being final physical qubit measurements
    Returns:
        List[Dict[str, float]]: the list of heavy output dictionaries of each of the quantum circuits.
    """
    simulable_circuits = deepcopy(qc_list)
    ideal_heavy_outputs: List[Dict[str, float]] = []
    ideal_simulator = Aer.get_backend("statevector_simulator")

    # Separate according to sorted indices
    circuit_batches = {
        k: [simulable_circuits[i] for i in sorted_qc_list_indices[k]] for k in sorted_qc_list_indices.keys()
    }

    for k in sorted(
        circuit_batches.keys(),
        key=lambda x: len(circuit_batches[x]),
        reverse=True,
    ):
        for qc in circuit_batches[k]:
            qc.remove_final_measurements()
            ideal_probabilities = ideal_simulator.run(qc).result().get_counts()
            heavy_projectors = heavy_projector(ideal_probabilities)
            if isinstance(heavy_projectors, list):
                ideal_heavy_outputs.extend(heavy_projectors)
            elif isinstance(heavy_projectors, dict):
                ideal_heavy_outputs.append(heavy_projectors)

    return ideal_heavy_outputs


def get_rem_hops(
    all_rem_quasidistro: List[List[QuasiCollection]], ideal_heavy_outputs: List[Dict[str, float]]
) -> List[float]:
    """Computes readout-error-mitigated heavy output probabilities.

    Args:
        all_rem_quasidistro (List[List[QuasiCollection]]): The list of lists of quasiprobability distributions.
        ideal_heavy_outputs (List[Dict[str, float]]): A list of the noiseless heavy output probability dictionaries.

    Returns:
        List[float]: A list of readout-error-mitigated heavy output probabilities.
    """
    qv_result_rem = []
    for rem_quasidistro, heavy in zip(all_rem_quasidistro, ideal_heavy_outputs):
        qv_result_rem += [expval(rem_quasidistro, heavy)]
    return qv_result_rem


def heavy_projector(probabilities: Dict[str, float]) -> Dict[str, float]:
    """
        Project (select) the samples from a given probability distribution onto heavy outputs.

    Args:
        probabilities (Dict[str, float]): A dictionary of bitstrings and associated probabilities.

    Returns:
        Dict[str, float]: The dictionary of heavy output bitstrings, all with weight 1.
    """
    median_prob = np.median(list(probabilities.values()))
    heavy_outputs = {k: 1.0 for k, v in probabilities.items() if v > median_prob}
    return heavy_outputs


def is_successful(
    heavy_output_probabilities: List[float],
    num_sigmas: int = 2,
) -> bool:
    """Check whether a QV benchmark returned heavy output results over the threshold, therefore being successful.

    This condition checks that the average of HOP is above the 2/3 threshold within the number of sigmas given in
    the configuration.

    Args:
        heavy_output_probabilities (List[float]): the HOP of all quantum circuits.
        num_sigmas (int): the number of sigmas to check
    Returns:
        bool: whether the QV benchmark was successful.
    """
    avg = np.mean(heavy_output_probabilities)
    std = (avg * (1 - avg) / len(heavy_output_probabilities)) ** 0.5

    return bool((avg - std * num_sigmas) > 2.0 / 3.0)


def plot_hop_threshold(
    qubits: List[int],
    depth: int,
    qv_result: List[float],
    qv_results_type: str,
    num_sigmas: int,
    backend_name: str,
    timestamp: str,
    in_volumetric: bool = False,
    plot_rem: bool = False,
) -> Tuple[str, Figure]:
    """Generate the figure representing each HOP, the average and the threshold.

    Args:
        qubits (List[int]): the list of qubit labels.
        depth (int): the depth of the QV circuit.
        qv_result (List[float]): the list of HOP.
        qv_results_type (str): whether results come from vanilla or DD execution.
        num_sigmas (int): the number of sigmas to plot.
        backend_name (str): the name of the backend.
        timestamp (str): the execution timestamp.
        in_volumetric (bool): whether the QV benchmark is being executed in the context of a volumetric benchmark.
            Defaults to False.
        plot_rem (bool): whether the plot corresponds to REM corrected data.

    Returns:
        str: the name of the figure.
        Figure: the figure.
    """
    cumul_hop = cumulative_hop(qv_result)
    cumul_std = cumulative_std(qv_result)

    fig = plt.figure()
    ax = plt.axes()

    plt.axhline(2.0 / 3.0, color="red", linestyle="dashed", label="Threshold")

    plt.scatter(
        np.arange(len(qv_result)),
        qv_result,
        marker=".",
        s=6,
        alpha=0.7,
        label="Individual HOP",
    )

    y_up: List[float] = [cumul_hop[i] + num_sigmas * cumul_std[i] for i in range(len(cumul_hop))]

    y_down: List[float] = [cumul_hop[i] - num_sigmas * cumul_std[i] for i in range(len(cumul_hop))]

    plt.fill_between(
        np.arange(len(qv_result)),
        y_up,
        y_down,
        color="b",
        alpha=0.125,
        label=rf"Cumulative {num_sigmas}$\sigma$",
    )

    plt.plot(cumul_hop, color=(0.0, 1.0, 0.5, 1.0), linewidth=2, label="Cumulative HOP")

    plt.ylim(min(qv_result), max(qv_result))
    ax.set_ylabel("Heavy Output Probability (HOP)")
    ax.set_xlabel("QV Circuit Samples (N)")
    plt.legend(loc="lower right")

    plt.margins(x=0, y=0)

    if in_volumetric:
        plt.title(
            f"Quantum Volume ({len(qubits)} qubits, {depth} depth)\nBackend: {backend_name} / {timestamp}",
            fontsize=9,
        )
        fig_name = f"{len(qubits)}_qubits_{depth}_depth"
    else:
        if plot_rem:
            plt.title(
                f"Quantum Volume ({qv_results_type}) with REM on {len(qubits)} qubits ({str(qubits)})\nBackend: {backend_name} / {timestamp}"
            )
            fig_name = f"{qv_results_type}_REM_{len(qubits)}_qubits_{str(qubits)}"
        else:
            plt.title(
                f"Quantum Volume ({qv_results_type}) on {len(qubits)} qubits ({str(qubits)})\nBackend: {backend_name} / {timestamp}"
            )
            fig_name = f"{qv_results_type}_{len(qubits)}_qubits_{str(qubits)}"

    plt.gcf().set_dpi(250)
    plt.close()

    return fig_name, fig


def qv_analysis(run: BenchmarkRunResult) -> BenchmarkAnalysisResult:
    """Analysis function for a Quantum Volume experiment

    Args:
        run (RunResult): A Quantum Volume experiment run for which analysis result is created
    Returns:
        AnalysisResult corresponding to Quantum Volume
    """

    plots = {}
    observations: list[BenchmarkObservation] = []
    dataset = run.dataset.copy(deep=True)
    backend_name = dataset.attrs["backend_name"]
    execution_timestamp = dataset.attrs["execution_timestamp"]
    num_circuits = dataset.attrs["num_circuits"]
    num_sigmas = dataset.attrs["num_sigmas"]

    physical_layout = dataset.attrs["physical_layout"]

    # Analyze the results for each qubit layout of the experiment dataset
    qubit_layouts = dataset.attrs["custom_qubits_array"]
    depth = {}
    qv_results_type = {}
    execution_results = {}
    ideal_heavy_outputs = {}
    rem = dataset.attrs["rem"]

    for qubits_idx, qubits in enumerate(qubit_layouts):
        qcvv_logger.info(f"Noiseless simulation and post-processing for layout {qubits}")
        # Retrieve counts
        execution_results[str(qubits)] = xrvariable_to_counts(dataset, str(qubits), num_circuits)

        # Retrieve other dataset values
        dataset_dictionary = dataset.attrs[qubits_idx]
        sorted_qc_list_indices = dataset_dictionary["sorted_qc_list_indices"]
        transpiled_circ_dataset = dataset.attrs["transpiled_circuits"][str(qubits)]
        transpiled_qc_list = []
        untranspiled_circ_dataset = dataset.attrs["untranspiled_circuits"][str(qubits)]
        qc_list = []
        for key in transpiled_circ_dataset:  # Keys (final layouts) are the same for transp/untransp
            transpiled_qc_list.extend(transpiled_circ_dataset[key])
            qc_list.extend(untranspiled_circ_dataset[key])

        qv_results_type[str(qubits)] = dataset_dictionary["qv_results_type"]
        depth[str(qubits)] = len(qubits)

        # Simulate the circuits and get the ideal heavy outputs
        ideal_heavy_outputs[str(qubits)] = get_ideal_heavy_outputs(qc_list, sorted_qc_list_indices)

        # Compute the HO probabilities
        qv_result = compute_heavy_output_probabilities(execution_results[str(qubits)], ideal_heavy_outputs[str(qubits)])

        observations.extend(
            [
                BenchmarkObservation(
                    name="average_heavy_output_probability",
                    value=cumulative_hop(qv_result)[-1],
                    uncertainty=cumulative_std(qv_result)[-1],
                    identifier=BenchmarkObservationIdentifier(qubits),
                ),
                BenchmarkObservation(
                    name="is_succesful",
                    value=is_successful(qv_result, num_sigmas),
                    identifier=BenchmarkObservationIdentifier(qubits),
                ),
                BenchmarkObservation(
                    name="QV_result",
                    value=2 ** len(qubits) if is_successful(qv_result) else 1,
                    identifier=BenchmarkObservationIdentifier(qubits),
                ),
            ]
        )

        dataset.attrs[qubits_idx].update(
            {
                "sorted_qc_list_indices": (sorted_qc_list_indices if rem or physical_layout == "batching" else None),
                "cumulative_average_heavy_output_probability": cumulative_hop(qv_result),
                "cumulative_stddev_heavy_output_probability": cumulative_std(qv_result),
                "heavy_output_probabilities": qv_result,
            }
        )

        fig_name, fig = plot_hop_threshold(
            qubits,
            depth[str(qubits)],
            qv_result,
            qv_results_type[str(qubits)],
            num_sigmas,
            backend_name,
            execution_timestamp,
            plot_rem=False,
        )
        plots[fig_name] = fig

    if not rem:
        return BenchmarkAnalysisResult(dataset=dataset, plots=plots, observations=observations)

    # When REM is set to True, do the post-processing with the adjusted quasi-probabilities
    mit_shots = dataset.attrs["mit_shots"]
    rem_quasidistros = dataset.attrs["REM_quasidistributions"]
    for qubits_idx, qubits in enumerate(qubit_layouts):
        qcvv_logger.info(f"REM post-processing for layout {qubits}")
        # Retrieve
        dataset_dictionary = dataset.attrs[qubits_idx]

        qcvv_logger.info(f"Applying REM with {mit_shots} shots")
        sorted_qc_list_indices = dataset_dictionary["sorted_qc_list_indices"]

        qv_result_rem = get_rem_hops(
            rem_quasidistros[f"REM_quasidist_{str(qubits)}"],
            ideal_heavy_outputs[str(qubits)],
        )

        dataset.attrs[qubits_idx].update(
            {
                "sorted_qc_list_indices": (sorted_qc_list_indices if physical_layout == "batching" else None),
                "REM_cumulative_average_heavy_output_probability": cumulative_hop(qv_result_rem),
                "REM_cumulative_stddev_heavy_output_probability": cumulative_std(qv_result_rem),
                "REM_heavy_output_probabilities": qv_result_rem,
            }
        )

        # UPDATE OBSERVATIONS
        observations.extend(
            [
                BenchmarkObservation(
                    name="REM_average_heavy_output_probability",
                    value=cumulative_hop(qv_result_rem)[-1],
                    uncertainty=cumulative_std(qv_result_rem)[-1],
                    identifier=BenchmarkObservationIdentifier(qubits),
                ),
                BenchmarkObservation(
                    name="REM_is_succesful",
                    value=is_successful(qv_result_rem, num_sigmas),
                    identifier=BenchmarkObservationIdentifier(qubits),
                ),
                BenchmarkObservation(
                    name="REM_QV_result",
                    value=2 ** len(qubits) if is_successful(qv_result_rem) else 1,
                    identifier=BenchmarkObservationIdentifier(qubits),
                ),
            ]
        )

        fig_name_rem, fig_rem = plot_hop_threshold(
            qubits,
            depth[str(qubits)],
            qv_result_rem,
            qv_results_type[str(qubits)],
            num_sigmas,
            backend_name,
            execution_timestamp,
            plot_rem=True,
        )
        plots[fig_name_rem] = fig_rem

    return BenchmarkAnalysisResult(dataset=dataset, plots=plots, observations=observations)


class QuantumVolumeBenchmark(Benchmark):
    """
    Quantum Volume reflects the deepest circuit a given number of qubits can execute with meaningful results.
    """

    analysis_function = staticmethod(qv_analysis)

    name: str = "quantum_volume"

    def __init__(self, backend_arg: IQMBackendBase | str, configuration: "QuantumVolumeConfiguration"):
        """Construct the QuantumVolumeBenchmark class.

        Args:
            backend_arg (IQMBackendBase): the backend to execute the benchmark on
            configuration (QuantumVolumeConfiguration): the configuration of the benchmark
        """
        super().__init__(backend_arg, configuration)

        self.backend_configuration_name = backend_arg if isinstance(backend_arg, str) else backend_arg.name

        self.num_circuits = configuration.num_circuits
        self.num_sigmas = configuration.num_sigmas
        self.choose_qubits_routine = configuration.choose_qubits_routine

        # self.mapomatic_num_qubits = configuration.mapomatic_num_qubits
        # self.mapomatic_num_layouts = configuration.mapomatic_num_layouts
        # self.mapomatic_qv_samples = configuration.mapomatic_qv_samples

        # self.dynamical_decoupling = configuration.dynamical_decoupling
        # self.dd_strategy = configuration.dd_strategy

        self.qiskit_optim_level = configuration.qiskit_optim_level
        self.optimize_sqg = configuration.optimize_sqg

        self.rem = configuration.rem
        self.mit_shots = configuration.mit_shots

        self.session_timestamp = strftime("%Y%m%d-%H%M%S")
        self.execution_timestamp = ""

        # The only difference between "custom" and "mapomatic" choose_qubits_routine is
        # that the custom qubit layout array is generated either by user-input or by mapomatic
        if self.choose_qubits_routine == "custom":
            if configuration.custom_qubits_array is None:
                raise ValueError(
                    "If the `choose_qubits_custom` routine is chosen, a `custom_qubits_array` must be specified in `QuantumVolumeConfiguration`."
                )
            self.custom_qubits_array = configuration.custom_qubits_array
        # elif self.choose_qubits_routine == "mapomatic":
        #    self.custom_qubits_array = self.get_mapomatic_average_qv_scores()
        else:
            raise ValueError("The `custom_qubits_array` variable must be either 'custom' or 'mapomatic'")

    def add_all_meta_to_dataset(self, dataset: xr.Dataset):
        """Adds all configuration metadata and circuits to the dataset variable

        Args:
            dataset (xr.Dataset): The xarray dataset
        """
        dataset.attrs["session_timestamp"] = self.session_timestamp
        dataset.attrs["execution_timestamp"] = self.execution_timestamp
        dataset.attrs["backend_configuration_name"] = self.backend_configuration_name
        dataset.attrs["backend_name"] = self.backend.name

        for key, value in self.configuration:
            if key == "benchmark":  # Avoid saving the class object
                dataset.attrs[key] = value.name
            else:
                dataset.attrs[key] = value
        # Defined outside configuration - if any

    @timeit
    def add_all_circuits_to_dataset(self, dataset: xr.Dataset):
        """Adds all generated circuits during execution to the dataset variable

        Args:
            dataset (xr.Dataset):  The xarray dataset

        Returns:

        """
        qcvv_logger.info(f"Adding all circuits to the dataset")
        for key, circuit in zip(
            ["transpiled_circuits", "untranspiled_circuits"], [self.transpiled_circuits, self.untranspiled_circuits]
        ):
            dictionary = {}
            for outer_key, outer_value in circuit.items():
                dictionary[str(outer_key)] = {
                    str(inner_key): inner_values for inner_key, inner_values in outer_value.items()
                }
            dataset.attrs[key] = dictionary

    # def get_mapomatic_average_qv_scores(self) -> List[List[int]]:
    #     """Estimate the average mapomatic scores for N quantum volume circuit samples
    #     Returns:
    #         List[List[object]]: the mapomatic layout scores sorted from the smallest mean*(1+std/sqrt(N)) cost score
    #     """
    #     qv_circ_samples = self.mapomatic_qv_samples
    #
    #     # Get calibration data once
    #     token = input("Trying to access IQM calibration data\nEnter your IQM token (without quote marks): ")
    #     calibration_data = get_calibration_fidelities(self.backend.architecture.name, token)
    #
    #     qcvv_logger.info(
    #         "Evaluating matching layouts for QV",
    #     )
    #
    #     all_scores: Dict[Tuple, List[float]] = {}
    #     for mapomatic_qv_sample in range(qv_circ_samples):
    #         qcvv_logger.info(
    #             f"Estimating layout costs for QV circuit sample {mapomatic_qv_sample+1}"
    #         )
    #         # Generate a representative quantum circuit sample
    #         qv_circ = get_circuit("quantum_volume", self.mapomatic_num_qubits)
    #         # Get all matching layouts
    #         layouts = matching_layouts(
    #             qv_circ,
    #             self.backend,
    #             self.backend.coupling_map,
    #             self.qiskit_optim_level,
    #         )
    #         # Evaluate all layouts
    #         scores = evaluate_costs(
    #             layouts,
    #             qv_circ,
    #             self.backend,
    #             calibration_data,
    #             self.qiskit_optim_level,
    #         )
    #         for score in scores:
    #             all_scores.setdefault(tuple(cast(list[int], score[0])), []).extend([cast(float, score[1])])
    #         qv_circ.clear()
    #
    #     # Consider mean and std of the cost scores as mean*(1+std/sqrt(N))
    #     all_scores_mean_plus_std = {
    #         s_k: np.mean(s_v) * (1 + np.std(s_v) / np.sqrt(qv_circ_samples)) for s_k, s_v in all_scores.items()
    #     }
    #     # Turn into an array ([[tuple_layout0, score0], [tuple_layout1,score1], ... ])
    #     sorted_scores_mean_plus_std = list(sorted(all_scores_mean_plus_std.items(), key=lambda y: y[1]))
    #     # Return indices to list type for layouts ([[list_layout0, score0], [list_layout1,score1], ... ])
    #     layout_meanscore = [[list(s[0]), s[1]] for s in sorted_scores_mean_plus_std]
    #
    #     mapomatic_qv_layouts = [x[0] for x in layout_meanscore[: self.mapomatic_num_layouts]]
    #     mapomatic_qv_costs = [x[1] for x in layout_meanscore[: self.mapomatic_num_layouts]]
    #
    #     qcvv_logger.info(
    #         f"Will execute QV on layouts {mapomatic_qv_layouts}, of mapomatic average costs {mapomatic_qv_costs}"
    #     )
    #
    #     return mapomatic_qv_layouts

    @staticmethod
    def generate_single_circuit(
        num_qubits: int,
        depth: Optional[int] = None,
        classical_permutation: bool = True,
    ) -> QuantumCircuit:
        """Generate a single QV quantum circuit, with measurements at the end.

        Args:
            num_qubits (int): number of qubits of the circuit
            depth (Optional[int]): The depth of the QV circuit. Defaults to None, which makes it equal to the number of
                qubits.
            classical_permutation (bool, optional): Whether permutations are classical, avoiding swapping layers.
                * Defaults to True.

        Returns:
            QuantumCircuit: the QV quantum circuit.
        """
        qc = QuantumVolume(num_qubits, depth=depth, classical_permutation=classical_permutation).decompose()
        qc.measure_all()
        return qc

    @timeit
    def generate_circuit_list(
        self,
        num_qubits: int,
        depth: Optional[int] = None,
        classical_permutations: bool = True,
    ) -> List[QuantumCircuit]:
        """Generate a list of QV quantum circuits, with measurements at the end.

        Args:
            num_qubits (int): the number of qubits of the circuits.
            depth (Optional[int]): The depth of the QV circuit. Defaults to None, which makes it equal to the number of qubits.
            classical_permutations (bool, optional): Whether permutations are classical, avoiding swapping layers.
                * Defaults to True.

        Returns:
            List[QuantumCircuit]: the list of QV quantum circuits.
        """
        qc_list = [
            self.generate_single_circuit(num_qubits, depth=depth, classical_permutation=classical_permutations)
            for _ in range(self.num_circuits)
        ]

        return qc_list

    def get_rem_quasidistro(
        self,
        sorted_transpiled_qc_list: Dict[Tuple, List[QuantumCircuit]],
        sorted_qc_list_indices: Dict[Tuple, List[int]],
        execution_results: List[Dict[str, int]],
        mit_shots: int,
    ) -> List[List[QuasiCollection]]:
        """Computes readout-error-mitigated quasiprobabilities.

        Args:
            sorted_transpiled_qc_list (Dict[Tuple, List[QuantumCircuit]]): A dictionary of lists of quantum circuits, indexed by qubiy layouts.
            sorted_qc_list_indices (Dict[Tuple, List[int]]): dictionary of indices (integers) corresponding to those in the original (untranspiled) list of circuits, with keys being final physical qubit measurements.
            execution_results (List[Dict[str, int]]): counts from execution of all quantum circuits.
            mit_shots (int): The number of measurement shots to estimate the readout calibration errors.
        Returns:
            A list of lists of quasiprobabilities.
        """
        all_rem_quasidistro = []
        for k in sorted(
            sorted_transpiled_qc_list.keys(),
            key=lambda x: len(sorted_transpiled_qc_list[x]),
            reverse=True,
        ):
            counts_corresp_to_circs_k = [execution_results[i] for i in sorted_qc_list_indices[k]]
            all_rem_quasidistro_batch_k, _ = apply_readout_error_mitigation(
                self.backend, sorted_transpiled_qc_list[k], counts_corresp_to_circs_k, mit_shots
            )
            all_rem_quasidistro += all_rem_quasidistro_batch_k

        return all_rem_quasidistro

    def submit_single_qv_job(
        self,
        backend: IQMBackendBase,
        qubits: Sequence[int],
        sorted_transpiled_qc_list: Dict[Tuple[int, ...], List[QuantumCircuit]],
    ) -> Dict[str, Any]:
        """
            Submit jobs for execution in the specified IQMBackend.
        Args:
            backend (IQMBackendBase): the IQM backend to submit the job.
            qubits (List[int]): the qubits to identify the submitted job.
            sorted_transpiled_qc_list (Dict[str, List[QuantumCircuit]]): qubits to submit jobs to.
        Returns:
            Dict with qubit layout, submitted job objects, type (vanilla/DD) and submission time.
        """
        time_submit = 0
        execution_jobs = None
        # qv_results_type = "dynamical decoupling"
        # if not self.dynamical_decoupling:
        qv_results_type = "vanilla"
        # Send to execute on backend
        execution_jobs, time_submit = submit_execute(
            sorted_transpiled_qc_list,
            backend,
            self.shots,
            self.calset_id,
            max_gates_per_batch=self.max_gates_per_batch,
        )
        # else:
        # DD IN DIQE VERSION PREVENTS SUBMITTING JOBS DYNAMICALLY:
        # I.E., IT AUTOMATICALLY RETRIEVES COUNTS
        # TODO: change this when job manager for DD in Pulla is updated! # pylint: disable=fixme
        # raise ValueError("Dynamical decoupling is not yet enabled in the new base")
        # qcvv_logger.info(
        #     f"Now executing {self.num_circuits} circuits with default strategy Dynamical Decoupling"
        # )
        # execution_results, time_retrieve = execute_with_dd(
        #     self.backend, transpiled_qc_list, self.shots, self.dd_strategy
        # )
        qv_results = {
            "qubits": qubits,
            "jobs": execution_jobs,
            "qv_results_type": qv_results_type,
            "time_submit": time_submit,
        }
        return qv_results

    def execute(self, backend: IQMBackendBase) -> xr.Dataset:
        """Executes the benchmark."""

        self.execution_timestamp = strftime("%Y%m%d-%H%M%S")

        dataset = xr.Dataset()
        self.add_all_meta_to_dataset(dataset)

        # Submit jobs for all qubit layouts first
        all_qv_jobs: List[Dict[str, Any]] = []
        time_circuit_generation = {}
        time_transpilation = {}
        time_batching = {}
        sorted_qc_list_indices = {}

        # Initialize the variable to contain the QV circuits of each layout
        self.untranspiled_circuits: Dict[str, Dict[Tuple, List[QuantumCircuit]]] = {}
        self.transpiled_circuits: Dict[str, Dict[Tuple, List[QuantumCircuit]]] = {}
        all_op_counts = {}

        for qubits in self.custom_qubits_array:  # NB: jobs will be submitted for qubit layouts in the specified order
            self.untranspiled_circuits[str(qubits)] = {}
            self.transpiled_circuits[str(qubits)] = {}
            num_qubits = len(qubits)
            depth = num_qubits
            qcvv_logger.info(f"Executing QV on qubits {qubits}")

            qc_list, time_circuit_generation[str(qubits)] = self.generate_circuit_list(num_qubits, depth=depth)
            qcvv_logger.info(f"Successfully generated all {self.num_circuits} circuits to be executed")
            # Set the coupling map
            coupling_map = set_coupling_map(qubits, backend, self.physical_layout)
            # Perform transpilation to backend
            qcvv_logger.info(f'Will transpile according to "{self.physical_layout}" physical layout')
            transpiled_qc_list, time_transpilation[str(qubits)] = perform_backend_transpilation(
                qc_list,
                backend=backend,
                qubits=qubits,
                coupling_map=coupling_map,
                qiskit_optim_level=self.qiskit_optim_level,
                optimize_sqg=self.optimize_sqg,
                routing_method=self.routing_method,
            )
            # Batching
            sorted_transpiled_qc_list: Dict[Tuple[int, ...], List[QuantumCircuit]] = {}
            time_batching[str(qubits)] = 0
            if self.physical_layout == "fixed":
                sorted_transpiled_qc_list = {tuple(qubits): transpiled_qc_list}
                sorted_qc_list_indices[str(qubits)] = {tuple(qubits): list(range(len(qc_list)))}
            elif self.physical_layout == "batching":
                # Sort circuits according to their final measurement mappings
                (
                    sorted_transpiled_qc_list,
                    sorted_qc_list_indices[str(qubits)],
                ), time_batching[
                    str(qubits)
                ] = sort_batches_by_final_layout(transpiled_qc_list)
            else:
                raise ValueError("physical_layout must either be \"fixed\" or \"batching\"")

            self.untranspiled_circuits[str(qubits)].update({tuple(qubits): qc_list})
            self.transpiled_circuits[str(qubits)].update(sorted_transpiled_qc_list)

            # Count operations
            all_op_counts[str(qubits)] = count_native_gates(backend, transpiled_qc_list)

            # Submit
            all_qv_jobs.append(self.submit_single_qv_job(backend, qubits, sorted_transpiled_qc_list))
            qcvv_logger.info(f"Job for layout {qubits} submitted successfully!")

        # Retrieve counts of jobs for all qubit layouts
        all_job_metadata = {}
        for job_idx, job_dict in enumerate(all_qv_jobs):
            qubits = job_dict["qubits"]
            # Retrieve counts
            execution_results, time_retrieve = retrieve_all_counts(job_dict["jobs"], str(qubits))
            # Retrieve all job meta data
            all_job_metadata = retrieve_all_job_metadata(job_dict["jobs"])

            # Export all to dataset
            dataset.attrs.update(
                {
                    job_idx: {
                        "qubits": qubits,
                        "qv_results_type": job_dict["qv_results_type"],
                        "time_circuit_generation": time_circuit_generation[str(qubits)],
                        "time_transpilation": time_transpilation[str(qubits)],
                        "time_batching": time_batching[str(qubits)],
                        "time_submit": job_dict["time_submit"],
                        "time_retrieve": time_retrieve,
                        "all_job_metadata": all_job_metadata,
                        "sorted_qc_list_indices": {
                            str(key): value for key, value in sorted_qc_list_indices[str(qubits)].items()
                        },
                        "operation_counts": all_op_counts[str(qubits)],
                    }
                }
            )

            qcvv_logger.info(f"Adding counts of {qubits} run to the dataset")
            dataset, _ = add_counts_to_dataset(execution_results, str(qubits), dataset)

        self.add_all_circuits_to_dataset(dataset)

        if self.rem:
            rem_quasidistros = {}
            for qubits in self.custom_qubits_array:
                exec_counts = xrvariable_to_counts(dataset, str(qubits), self.num_circuits)
                rem_quasidistros[f"REM_quasidist_{str(qubits)}"] = self.get_rem_quasidistro(
                    self.transpiled_circuits[str(qubits)],
                    sorted_qc_list_indices[str(qubits)],
                    exec_counts,
                    self.mit_shots,
                )
            dataset.attrs.update({"REM_quasidistributions": rem_quasidistros})

        qcvv_logger.info(f"QV experiment execution concluded !")
        return dataset


class QuantumVolumeConfiguration(BenchmarkConfigurationBase):
    """Quantum Volume configuration.

    Attributes:
        benchmark (Type[Benchmark]): QuantumVolumeBenchmark
        num_circuits (int): The number of circuits to use.
                            * Should be at least 100 for a meaningful QV experiment.
        num_sigmas (int): The number of sample standard deviations to consider with for the threshold criteria.
                            * Default by consensus is 2
        choose_qubits_routine (Literal["custom"]): The routine to select qubit layouts.
                            * Default is "custom".
        custom_qubits_array (Optional[Sequence[Sequence[int]]]): The physical qubit layouts to perform the benchmark on.
                            * Default is [[0, 2]].
        qiskit_optim_level (int): The Qiskit transpilation optimization level.
                            * Default is 3.
        optimize_sqg (bool): Whether Single Qubit Gate Optimization is performed upon transpilation.
                            * Default is True.
        routing_method (Literal["basic", "lookahead", "stochastic", "sabre", "none"]): The Qiskit transpilation routing method to use.
                            * Default is "sabre".
        physical_layout (Literal["fixed", "batching"]): Whether the coupling map is restricted to qubits in the input layout or not.
                            - "fixed": Restricts the coupling map to only the specified qubits.
                            - "batching": Considers the full coupling map of the backend and circuit execution is batched per final layout.
                            * Default is "fixed"
        rem (bool): Whether Readout Error Mitigation is applied in post-processing.
                    When set to True, both results (readout-unmitigated and -mitigated) are produced.
                            - Default is True.
        mit_shots (int): The measurement shots to use for readout calibration.
                            * Default is 1_000.
    """

    benchmark: Type[Benchmark] = QuantumVolumeBenchmark
    num_circuits: int
    num_sigmas: int = 2
    choose_qubits_routine: Literal["custom"] = "custom"
    custom_qubits_array: Sequence[Sequence[int]]
    qiskit_optim_level: int = 3
    optimize_sqg: bool = True
    rem: bool = True
    mit_shots: int = 1_000



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Generic Benchmark class
"""

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Dict, Literal, Optional, OrderedDict, Type

from matplotlib.figure import Figure
from pydantic import BaseModel

from iqm.qiskit_iqm.iqm_backend import IQMBackendBase


class BenchmarkBase(ABC):
    """
    The base implementation of all benchmarks, from which they inherit.
    """

    def __init__(
        self,
        backend: IQMBackendBase,
        configuration: "BenchmarkConfigurationBase",
    ):
        """Construct the BenchmarkBase class.

        Args:
            backend (IQMBackendBase): the backend to execute the benchmark on
            configuration (BenchmarkConfigurationBase): the configuration of the benchmark
        """
        self.backend = backend
        self.configuration = configuration
        self.serializable_configuration = deepcopy(self.configuration)
        self.serializable_configuration.benchmark = self.name()

        self.shots = self.configuration.shots
        self.calset_id = self.configuration.calset_id
        self.max_gates_per_batch = self.configuration.max_gates_per_batch

        self.routing_method = self.configuration.routing_method
        self.physical_layout = self.configuration.physical_layout

        self.raw_data: Dict = {}
        self.job_meta: Dict = {}
        self.results: Dict = {}
        self.raw_results: Dict = {}

        self.untranspiled_circuits: Dict[str, Dict[int, list]] = {}
        self.transpiled_circuits: Dict[str, Dict[int, list]] = {}

        self.figures: Dict[str, Figure] = {}

    @classmethod
    @abstractmethod
    def name(cls):
        """Return the name of the benchmark."""

    @abstractmethod
    def execute_full_benchmark(self):
        """Execute the full benchmark on the given backend."""

    @staticmethod
    def check_requirements(all_benchmarks: OrderedDict[str, "BenchmarkBase"]) -> OrderedDict[str, "BenchmarkBase"]:
        """Check whether the requirements for the benchmark are met, returning a valid benchmark dictionary."""
        return all_benchmarks

    def generate_requirements(self, all_benchmarks: OrderedDict[str, "BenchmarkBase"]) -> None:
        """Generate the required attributes for execution."""


class BenchmarkConfigurationBase(BaseModel):
    """Benchmark configuration base.

    Attributes:
        benchmark (Type[BenchmarkBase]): the benchmark configuration.
        shots (int): the number of shots to use in circuit execution.
                * Default for all benchmarks is 2**8.
        max_gates_per_batch (Optional[int]): the maximum number of gates per circuit batch.
                * Default for all benchmarks is None.
        calset_id (Optional[str]): the calibration ID to use in circuit execution.
                * Default for all benchmarks is None (uses last available calibration ID).
        routing_method (Literal["basic", "lookahead", "stochastic", "sabre", "none"]): the Qiskit routing method to use in transpilation.
                * Default for all benchmarks is "sabre".
        physical_layout (Literal["fixed", "batching"]): whether physical layout is constrained during transpilation to selected physical qubits.
                - "fixed": physical layout is constrained during transpilation to the selected initial physical qubits.
                - "batching": physical layout is allowed to use any other physical qubits, and circuits are batched according to final measured qubits.
                * Default for all benchmarks is "fixed".
    """

    benchmark: Type[BenchmarkBase]
    shots: int = 2**8
    max_gates_per_batch: Optional[int] = None
    calset_id: Optional[str] = None
    routing_method: Literal["basic", "lookahead", "stochastic", "sabre", "none"] = "sabre"
    physical_layout: Literal["fixed", "batching"] = "fixed"



"""
Generation of error measures and result tables
"""

from argparse import Namespace
import csv
import os

import numpy as np
import numpy.linalg as la
import pandas as pd
from pygsti.algorithms import gaugeopt_to_target
from pygsti.baseobjs import Basis
from pygsti.models import gaugegroup
from pygsti.report.reportables import entanglement_fidelity
from pygsti.tools import change_basis
from pygsti.tools.optools import compute_povm_map
from qiskit.quantum_info import SuperOp
from qiskit.quantum_info.operators.measures import diamond_norm
from scipy.linalg import logm
from scipy.optimize import linear_sum_assignment, minimize

from mGST import additional_fns, algorithm, compatibility, low_level_jit


def min_spectral_distance(X1, X2):
    """Computes the average absolute distance between the eigenvlues of two matrices
    The matrices are first diagonalized, then the eigenvalues are matched such that the average
    distance of matched eigenvalue pairs is minimal.
    The big advantage of this distance metric is that it is gauge invariant and it can thus be used
    to if the reconstructed gates are similar to the target gates before any gauge optimization.

    Parameters
    ----------
    X1: numpy array
        The first matrix
    X2: numpy array
        The second matrix

    Returns
    -------
    dist: float
        The minimal distance
    """
    r = X1.shape[0]
    eigs = la.eig(X1)[0]
    eigs_t = la.eig(X2)[0]
    cost_matrix = np.array([[np.abs(eigs[i] - eigs_t[j]) for i in range(r)] for j in range(r)])
    row_ind, col_ind = linear_sum_assignment(cost_matrix)
    normalization = np.abs(eigs).sum()
    dist = cost_matrix[row_ind, col_ind].sum() / normalization
    return dist


def MVE_data(X, E, rho, J, y):
    """Mean varation error between measured outcomes and predicted outcomes from a gate set

    Parameters
    ----------
    X : numpy array
        Gate set
    E : numpy array
        POVM
    rho : numpy array
        Initial state
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence

    Returns
    -------
    dist : float
        Mean variation error
    max_dist : float
        Maximal varaition error

    Notes:
        For each sequence the total variation error of the two probability distribution
        over the POVM elements is computed. Afterwards the meean over these total
        variation errors is returned.
    """
    m = y.shape[1]
    n_povm = y.shape[0]
    dist: float = 0
    max_dist: float = 0
    curr: float = 0
    for i in range(m):
        j = J[i]
        C = low_level_jit.contract(X, j)
        curr = 0
        for k in range(n_povm):
            y_model = E[k].conj() @ C @ rho
            curr += np.abs(y_model - y[k, i])
        curr = curr / 2
        dist += curr
        max_dist = max(max_dist, curr)
    return dist / m, max_dist


def gauge_opt(X, E, rho, target_mdl, weights):
    """Performs pyGSti gauge optimization to target model

    Parameters
    ----------
    X : numpy array
        Gate set
    E : numpy array
        POVM
    rho : numpy array
        Initial state
    target_mdl : pygsti model object
        A model containing the target gate set
    weights : dict[str: float]
        A dictionary with keys being gate labels or "spam" and corresponding values being the
        weight of each gate in the gauge optimization.
        Example for uniform weights: dict({"G%i"%i:1  for i in range(d)}, **{"spam":1})

    Returns
    -------
    X_opt, E_opt, rho_opt: Numpy arrays
        The gauge optimized gates and SPAM arrays
    """
    mdl = compatibility.arrays_to_pygsti_model(X, E, rho, basis="std")
    X_t, E_t, rho_t = compatibility.pygsti_model_to_arrays(target_mdl, basis="std")
    target_mdl = compatibility.arrays_to_pygsti_model(X_t, E_t, rho_t, basis="std")  # For consistent gate labels

    gauge_optimized_mdl = gaugeopt_to_target(
        mdl,
        target_mdl,
        gauge_group=gaugegroup.UnitaryGaugeGroup(target_mdl.state_space, basis="pp"),
        item_weights=weights,
    )
    return compatibility.pygsti_model_to_arrays(gauge_optimized_mdl, basis="std")


def report(X, E, rho, J, y, target_mdl, gate_labels):
    """Generation of pandas dataframes with gate and SPAM quality measures
    The resutls can be converted to .tex tables or other formats to be used for GST reports

    Parameters
    ----------
    X : numpy array
        Gate set
    E : numpy array
        POVM
    rho : numpy array
        Initial state
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    target_mdl : pygsti model object
        A model containing the target gate set
    gate_labels : list[str]
        A list of names for the gates in X

    Returns
    -------
        df_g : Pandas DataFrame
            DataFrame of gate quality measures
        df_o : Pandas DataFrame
            DataFrame of all other quality/error measures
        s_g : Pandas DataFrame.style object
        s_o : Pandas DataFrame.style object
    """
    pdim = int(np.sqrt(rho.shape[0]))
    X_t, E_t, rho_t = compatibility.pygsti_model_to_arrays(target_mdl, basis="std")
    target_mdl = compatibility.arrays_to_pygsti_model(X_t, E_t, rho_t, basis="std")  # For consistent gate labels

    gauge_optimized_mdl = compatibility.arrays_to_pygsti_model(X, E, rho, basis="std")
    E_map = compute_povm_map(gauge_optimized_mdl, "Mdefault")
    E_map_t = compute_povm_map(target_mdl, "Mdefault")

    final_objf = low_level_jit.objf(X, E, rho, J, y)
    MVE = MVE_data(X, E, rho, J, y)[0]
    MVE_target = MVE_data(X_t, E_t, rho_t, J, y)[0]

    povm_dd = diamond_norm(SuperOp(E_map) - SuperOp(E_map_t)) / 2
    rho_td = la.norm(rho.reshape((pdim, pdim)) - rho_t.reshape((pdim, pdim)), ord="nuc") / 2
    F_avg = compatibility.average_gate_fidelities(gauge_optimized_mdl, target_mdl, pdim, basis_string="pp")
    DD = [diamond_norm(SuperOp(X[i]) - SuperOp(X_t[i])) / 2 for i in range(len(X))]
    min_spectral_dists = [min_spectral_distance(X[i], X_t[i]) for i in range(X.shape[0])]

    df_g = pd.DataFrame({"F_avg": F_avg, "Diamond distances": DD, "Min. Spectral distances": min_spectral_dists})
    df_o = pd.DataFrame(
        {
            "Final cost function value": final_objf,
            "Mean total variation dist. to data": MVE,
            "Mean total variation dist. target to data": MVE_target,
            "POVM - Meas. map diamond dist.": povm_dd,
            "State - Trace dist.": rho_td,
        },
        index=[0],
    )
    df_g.rename(index=gate_labels, inplace=True)
    df_o.rename(index={0: ""}, inplace=True)

    return df_g, df_o


def quick_report(X, E, rho, J, y, target_mdl, gate_labels=None):
    """Generation of pandas dataframes with gate and SPAM quality measures
    The quick report is intended to check on a GST estimate with fast to compute measures
    (no diamond distance) to get a first picture and check whether mGST and the gauge optimization
    produce meaningful results.

    Parameters
    ----------
    X : numpy array
        Gate set
    E : numpy array
        POVM
    rho : numpy array
        Initial state
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    target_mdl : pygsti model object
        A model containing the target gate set
    gate_labels : list[str]
        A list of names for the gates in X

    Returns
    -------
        df_g : Pandas DataFrame
            DataFrame of gate quality measures
        df_o : Pandas DataFrame
            DataFrame of all other quality/error measures
        s_g : Pandas DataFrame.style object
        s_o : Pandas DataFrame.style object
    """
    pdim = int(np.sqrt(rho.shape[0]))
    d = X.shape[0]
    if not gate_labels:
        gate_labels = {i: f"Gate %i" % i for i in range(d)}

    X_t, E_t, rho_t = compatibility.pygsti_model_to_arrays(target_mdl, basis="std")
    target_mdl = compatibility.arrays_to_pygsti_model(X_t, E_t, rho_t, basis="std")  # For consistent gate labels

    gauge_optimized_mdl = compatibility.arrays_to_pygsti_model(X, E, rho, basis="std")

    final_objf = low_level_jit.objf(X, E, rho, J, y)
    MVE = MVE_data(X, E, rho, J, y)[0]
    MVE_target = MVE_data(X_t, E_t, rho_t, J, y)[0]

    E_map = compute_povm_map(gauge_optimized_mdl, "Mdefault")
    E_map_t = compute_povm_map(target_mdl, "Mdefault")
    povm_dd = diamond_norm(SuperOp(E_map) - SuperOp(E_map_t)) / 2

    rho_td = la.norm(rho.reshape((pdim, pdim)) - rho_t.reshape((pdim, pdim)), ord="nuc") / 2
    F_avg = compatibility.average_gate_fidelities(gauge_optimized_mdl, target_mdl, pdim, basis_string="pp")
    min_spectral_dists = [min_spectral_distance(X[i], X_t[i]) for i in range(X.shape[0])]

    df_g = pd.DataFrame({"F_avg": F_avg, "Min. Spectral distances": min_spectral_dists})
    df_o = pd.DataFrame(
        {
            "Final cost function": final_objf,
            "Mean TVD estimate-data": MVE,
            "Mean TVD target-data": MVE_target,
            "SPAM error:": rho_td + povm_dd,
        },
        index=[0],
    )
    df_g.rename(index=gate_labels, inplace=True)
    df_o.rename(index={0: ""}, inplace=True)

    return df_g, df_o


def compute_angles_axes(U_set, alternative_phase=False):
    """Takes the matrix logarithm of the given unitaries and returns the Hamiltonian parameters
    The parametrization is U = exp(-i pi H/2), i.e. H = i log(U)*2/pi

    Parameters
    ----------
    U_set: list[numpy array]
        A list contining unitary matrices
    alternative_phase: bool
        Whether an attempt should be made to report more intuitive rotations,
        for example a rotation of 3 pi/2 around the -X axis would be turned into
        a pi/2 rotation around the X-axis.
    Returns
    -------
    angles: list[float]
        The rotation angle on the Bloch sphere for all gates
    axes : list[numpy array]
        The normalized rotation axes in the Pauli basis for all gates
    pauli_coeffs : list[numpy array]
        The full list of Pauli basis coefficients of the Hamiltonian for all gates

    Notes: sqrt(pdim) factor is due to Pauli basis normalization
    """
    d = U_set.shape[0]
    pdim = U_set.shape[1]
    angles = []
    axes = []
    pp_vecs = []
    for i in range(d):
        H = 1j * logm(U_set[i])
        pp_vec = change_basis(H.reshape(-1), "std", "pp")
        original_phase = la.norm(pp_vec[1:]) * 2 / np.sqrt(pdim)
        if alternative_phase and (-np.min(pp_vec) > np.max(pp_vec)) and original_phase > np.pi:
            alt_phase = (-original_phase + 2 * np.pi) % (2 * np.pi)
            pp_vec = -pp_vec
        else:
            alt_phase = original_phase
        angles.append(alt_phase / np.pi)
        axes.append(pp_vec[1:] / la.norm(pp_vec[1:]))
        pp_vecs.append(pp_vec)
    pauli_coeffs = np.array(pp_vecs) / np.sqrt(pdim) / np.pi * 2
    return angles, axes, pauli_coeffs


def compute_sparsest_Pauli_Hamiltonian(U_set):
    """Takes the matrix logarithms of the given unitaries and returns sparsest Hamiltonian parameters in Pauli basis
    The parametrization is U = exp(-i pi H/2), i.e. H = i log(U)*2/pi.
    Different branches in the matrix logarithm lead to different Hamiltonians. This function optimizes over
    combinations of adding 2*pi to different eigenvalues, in order arrive at the branch with the Hamiltonian
    whose Pauli basis representation is the most sparse.


    Parameters
    ----------
    U_set  : list[numpy array]
        A list contining unitary matrices

    Returns
    -------
    pauli_coeffs : list[numpy array]
        The full list of Pauli basis coefficients of the Hamiltonian for all gates

    Notes: sqrt(pdim) factor is due to Pauli basis normalization
    """
    pdim = U_set.shape[1]
    pp_vecs = []

    for U in U_set:
        evals, evecs = np.linalg.eig(U)
        Pauli_norms = []
        for i in range(2**pdim):
            bits = low_level_jit.local_basis(i, 2, pdim)
            evals_new = 1j * np.log(evals) + 2 * np.pi * bits
            H_new = evecs @ np.diag(evals_new) @ evecs.T.conj()
            pp_vec = change_basis(H_new.reshape(-1), "std", "pp")
            Pauli_norms.append(np.linalg.norm(pp_vec, ord=1))
        opt_bits = low_level_jit.local_basis(np.argsort(Pauli_norms)[0], 2, pdim)
        evals_opt = 1j * np.log(evals) + 2 * np.pi * opt_bits
        H_opt = evecs @ np.diag(evals_opt) @ evecs.T.conj()
        pp_vecs.append(change_basis(H_opt.reshape(-1), "std", "pp"))
    pauli_coeffs = np.array(pp_vecs) / np.sqrt(pdim) / np.pi * 2
    return pauli_coeffs


def phase_err(angle, U, U_t):
    """Computes norm between two input unitaries after a global phase is added to one of them

    Parameters
    ----------
    angle : float
        The global phase angle (in rad)
    U : numpy array
        The first unitary matrix
    U_t : numpy array
        The second unitary matrix (typically the a target gate)

    Returns
    -------
    norm : floar
        The norm of the difference
    """
    return la.norm(np.exp(1j * angle) * U - U_t)


def phase_opt(X, K_t):
    """Return rK = 1 gate set with global phase fitting matching to target gate set

    Parameters
    ----------
    X: 3D numpy array
        Array where CPT superoperators are stacked along the first axis.
        These should correspond to rK = 1 gates for the outputs to be meaningful.
    K_t: 4D numpy array
        Array of target gate Kraus operators
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.

    Returns
    -------
    K_opt: 4D numpy array
        Array of Kraus operators with mathed global phase
    """
    d = X.shape[0]
    r = X.shape[1]
    pdim = int(np.sqrt(r))
    K = additional_fns.Kraus_rep(X, d, pdim, 1).reshape((d, pdim, pdim))
    K_t = K_t.reshape(d, pdim, pdim)
    K_opt = np.zeros(K.shape).astype(complex)
    for i in range(d):
        angle_opt = minimize(phase_err, 1e-9, args=(K[i], K_t[i]), method="COBYLA").x
        K_opt[i] = K[i] * np.exp(1j * angle_opt)
    return K_opt


def eff_depol_params_agf(X_opt_pp):
    """Computes the average gate fidelities to the completely depolarizing channel

    Parameters
    ----------
    X_opt_pp: 3D numpy array
        Array where CPT superoperators in Pauli basis are stacked along the first axis.

    Returns
    -------
    ent_fids : list[float]
        List of average gate fidelities wrt. the depol. channel corresponding to the gates in X_opt_pp.
    """
    r = X_opt_pp.shape[1]
    pdim = np.sqrt(r)
    ent_fids = []
    basis = Basis.cast("pp", r)
    K_depol = additional_fns.depol(int(np.sqrt(r)), 1)
    X_depol = np.einsum("jkl,jnm -> knlm", K_depol, K_depol.conj()).reshape(r, r)
    for i in range(X_opt_pp.shape[0]):
        ent_fids.append(entanglement_fidelity(X_opt_pp[i], change_basis(X_depol, "std", "pp"), basis))
    return (pdim * np.array(ent_fids) + 1) / (pdim + 1)


def unitarities(X_opt_pp):
    """Computes the unitarities of all gates in the gate set

    Parameters
    ----------
    X_opt_pp : 3D numpy array
        Array where CPT superoperators in Pauli basis are stacked along the first axis.

    Returns
    -------
    unitarities : list[float]
        List of unitarities for the gates in X_opt_pp.
    """
    # Definition: Proposition 1 of https://arxiv.org/pdf/1503.07865.pdf
    pdim = int(np.sqrt(X_opt_pp.shape[1]))
    E_u = X_opt_pp[:, 1:, 1:]  # unital submatrices
    unitarity_list = np.real(np.einsum("ijk, ijk -> i", E_u.conj(), E_u) / (pdim**2 - 1))
    return unitarity_list


# Spectrum of the Choi matrix
def generate_Choi_EV_table(X_opt, n_evals, gate_labels, filename=None):
    """Outputs a .tex document containing a table with the larges eigenvalues of the Choi matrix for each gate

    Parameters
    ----------
    X_opt : 3D numpy array
        Array where CPT superoperators in standard basis are stacked along the first axis.
    n_evals : int
        Number of eigenvalues to be returned
    gate_labels : list[int: str]
        The names of gates in the gate set
    filename :
        The file name of the output .tex file
    """
    d, r, _ = X_opt.shape
    pdim = int(np.sqrt(r))
    Choi_evals_result = np.zeros((d, r))
    X_choi = X_opt.reshape(d, pdim, pdim, pdim, pdim)
    X_choi = np.einsum("ijklm->iljmk", X_choi).reshape((d, pdim**2, pdim**2))
    for j in range(d):
        Choi_evals_result[j, :] = np.sort(np.abs(la.eig(X_choi[j])[0]))[::-1]
    Choi_evals_normalized = np.einsum("ij,i -> ij", Choi_evals_result, 1 / la.norm(Choi_evals_result, axis=1, ord=1))

    df_g_evals = pd.DataFrame(Choi_evals_normalized)
    df_g_evals.rename(index=gate_labels, inplace=True)
    if filename:
        df_g_evals.style.to_latex(
            filename + ".tex",
            column_format="c|*{%i}{c}" % n_evals,
            position_float="centering",
            hrules=True,
            caption="Eigenvalues of the Choi state",
            position="h!",
        )
    return df_g_evals


def dephasing_dist(prob_vec, X_pp):
    """Returns the distance between a given channel and a local dephasing channel with given probabilities

    Parameters
    ----------
    prob_vec : list[float]
        A list of dephasing probabilities
    X_pp : 3D numpy array
        Array where CPT superoperators in Pauli basis are stacked along the first axis.

    Returns
    -------
    norm : float
        The norm between the channel difference

    """
    X_deph = additional_fns.local_dephasing_pp(prob_vec)
    return la.norm(X_pp - X_deph)


def dephasing_probabilities_2q(X_opt_pp, X_ideal_pp):
    """Determines the local dephasing channel parameters which best describe the noise model
    Works for two qubit gates only

    Parameters
    ----------
    X_opt_pp : 3D numpy array
        Array where reconstructed CPT superoperators in Pauli basis are stacked along the first axis.
    X_opt_pp : 3D numpy array
        Array where target gate CPT superoperators in Pauli basis are stacked along the first axis.

    Returns
    -------
    dephasing_probs : list[float]
        The two best fit dephasing probabilities
    """
    dephasing_probs = []
    for i in range(X_opt_pp.shape[0]):
        dephasing_probs.append(minimize(dephasing_dist, [0.1, 0.1], args=X_opt_pp[i] @ la.inv(X_ideal_pp[i])).x)
    return dephasing_probs


def bootstrap_errors(K, X, E, rho, mGST_args, bootstrap_samples, weights, gate_labels, target_mdl, parametric=False):
    """Resamples circuit outcomes a number of times and computes GST estimates for each repetition
    All results are then returned in order to compute bootstrap-error bars for GST estimates.
    Parametric bootstrapping uses the estimated gate set to create a newly sampled data set.
    Non-parametric bootstrapping uses the initial dataset and resamples according to the
    corresp. outcome probabilities.
    Each bootstrap run is initialized with the estimated gate set in order to save processing time.

    Parameters
    ----------
    K : numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.
    X : 3D numpy array
        Array where reconstructed CPT superoperators in standard basis are stacked along the first axis.
    E : numpy array
        Current POVM estimate
    rho : numpy array
        Current initial state estimate
    mGST_args : dict[str: misc]
        Arguments with which the run_mGST function was called
    bootstrap_samples : int
        Number of bootstrapping repretitions
    weights : dict[str: float]
        Gate weights used for gauge optimization
    gate_labels : list[int: str]
        The names of gates in the gate set
    target_mdl : pygsti model object
        The target gate set
    parametric : bool
        If set to True, parametric bootstrapping is used, else non-parametric bootstrapping. Default: False

    Returns
    -------
    X_array : numpy array
        Array containing all estimated gate tensors of different bootstrapping repretitions along first axis
    E_array : numpy array
        Array containing all estimated POVM tensors of different bootstrapping repretitions along first axis
    rho_array : numpy array
        Array containing all estimated initial states of different bootstrapping repretitions along first axis
    df_g_array : numpy array
        Contains gate quality measures of bootstrapping repetitions
    df_o_array : numpy array
        Contains SPAM and other quality measures of bootstrapping repetitions

    """
    ns = Namespace(**mGST_args)
    if parametric:
        y = np.real(
            np.array([[E[i].conj() @ low_level_jit.contract(X, j) @ rho for j in ns.J] for i in range(ns.n_povm)])
        )
    else:
        y = ns.y
    X_array = np.zeros((bootstrap_samples, *X.shape)).astype(complex)
    E_array = np.zeros((bootstrap_samples, *E.shape)).astype(complex)
    rho_array = np.zeros((bootstrap_samples, *rho.shape)).astype(complex)
    df_g_list = []
    df_o_list = []

    for i in range(bootstrap_samples):
        y_sampled = additional_fns.sampled_measurements(y, ns.meas_samples).copy()
        _, X_, E_, rho_, _ = algorithm.run_mGST(
            y_sampled,
            ns.J,
            ns.l,
            ns.d,
            ns.r,
            ns.rK,
            ns.n_povm,
            ns.bsize,
            ns.meas_samples,
            method=ns.method,
            max_inits=ns.max_inits,
            max_iter=0,
            final_iter=ns.final_iter,
            threshold_multiplier=ns.threshold_multiplier,
            target_rel_prec=ns.target_rel_prec,
            init=[K, E, rho],
            testing=False,
        )

        X_opt, E_opt, rho_opt = gauge_opt(X_, E_, rho_, target_mdl, weights)
        df_g, df_o = report(X_opt, E_opt, rho_opt, ns.J, y_sampled, target_mdl, gate_labels)
        df_g_list.append(df_g.values)
        df_o_list.append(df_o.values)

        X_opt_pp, E_opt_pp, rho_opt_pp = compatibility.std2pp(X_opt, E_opt, rho_opt)

        X_array[i, :] = X_opt_pp
        E_array[i, :] = E_opt_pp
        rho_array[i, :] = rho_opt_pp

    return (X_array, E_array, rho_array, np.array(df_g_list), np.array(df_o_list))


def job_counts_to_mGST_format(self, result_dict):
    """Turns the dictionary of outcomes obtained from qiskit backend
        into the format which is used in mGST

    Parameters
    ----------
    result_dict: (dict of str: int)

    Returns
    -------
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence

    """
    basis_dict_list = []
    for result in result_dict:
        # Translate dictionary entries of bitstring on the full system to the decimal representation of bitstrings on the active qubits
        basis_dict = {entry: int("".join([entry[::-1][i] for i in self.qubits][::-1]), 2) for entry in result}
        # Sort by index:
        basis_dict = dict(sorted(basis_dict.items(), key=lambda item: item[1]))
        basis_dict_list.append(basis_dict)
    y = []
    for i in range(len(result_dict)):
        row = [result_dict[i][key] for key in basis_dict_list[i]]
        if len(row) < self.num_povm:
            missing_entries = list(np.arange(self.num_povm))
            for given_entry in basis_dict_list[i].values():
                missing_entries.remove(given_entry)
            for missing_entry in missing_entries:
                row.insert(missing_entry, 0)  # 0 measurement outcomes in not recorded entry
        y.append(row / np.sum(row))
    y = np.array(y).T
    return y


def save_var_latex(key, value):
    """Saves variables in data file to be read in latex document
    Credit to https://stackoverflow.com/a/66620671

    Parameters
    ----------
    key : str
        The name of the variable
    value : misc
        The variable content, could be a string as in the name of the experiment, or the number of circuits run (int),
        or any other python variable that needs to be transfered
    """

    dict_var = {}

    file_path = os.path.join(os.getcwd(), "report/latex_vars.dat")

    try:
        with open(file_path, newline="", encoding="ASCII") as file:
            reader = csv.reader(file)
            for row in reader:
                dict_var[row[0]] = row[1]
    except FileNotFoundError:
        pass

    dict_var[key] = value

    with open(file_path, "w", encoding="ASCII") as f:
        for key_, value_ in dict_var:
            f.write(f"{key_},{value_}\n")


def number_to_str(number, uncertainty=None, precision=3):
    """Formats a floating point number to a string with the given precision.

    Parameters:
    number (float): The floating point number to format.
    uncertainty (tuple): The upper and lower values of the confidence interval
    precision (int): The number of decimal places to include in the formatted string.

    Returns:
    str: The formatted floating point number as a string.
    """
    if uncertainty is None:
        return f"{number:.{precision}f}"

    return f"{number:.{precision}f} [{uncertainty[1]:.{precision}f},{uncertainty[0]:.{precision}f}]"



"""
Functions related to optimization on manifolds
"""

import numpy as np
from scipy.linalg import eigh

from mGST.low_level_jit import ddM, dK_dMdM, objf


def eigy_expm(A):
    """Custom Matrix exponential using the eigendecomposition of numpy.linalg

    Parameters
    ----------
    A : numpy array
        Matrix to be exponentiated

    Returns
    -------
    M: numpy array
        Matrix exponential of A
    """
    vals, vects = np.linalg.eig(A)
    return np.einsum("...ik, ...k, ...kj -> ...ij", vects, np.exp(vals), np.linalg.inv(vects))


def tangent_proj(K, Z, d, rK):
    """Projection onto the local tangent space

    Parameters
    ----------
    K : numpy array
        Current position
    Z : numpy array
        Element of the ambient space to be projected onto the tangent space at K
    d : int
        Number of matrices which are projected (i.e. number of gates in the gate set)
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension

    Returns
    -------
    G: numpy array
        Projection of Z onto the tangent space of the Stiefel manifold at position K

    Notes:
        The projection is done with respect to the canonical metrik.
    """
    pdim = K.shape[2]
    n = pdim * rK
    K = K.reshape(d, n, pdim)
    Z = Z.reshape(d, n, pdim)
    G = np.ascontiguousarray(np.zeros((d, n, pdim)).astype(np.complex128))
    for i in range(d):
        G[i] += Z[i] - (K[i] @ K[i].T.conj() @ Z[i] + K[i] @ Z[i].T.conj() @ K[i]) / 2
    return G


def update_K_geodesic(K, H, a):
    """Compute a new point on the geodesic

    G = np.ascontiguousarray(np.zeros((d,n,pdim)).astype(np.complex128))

    Compute a new point on the geodesic

    Parameters
    ----------
    K : numpy array
        Current position
    H : numpy array
        Element of the tangent space at K and local direction of the geodesic
    a : float
        Geodesic curve parameter

    Returns
    -------
    K_new: numpy array
        New position given by K_new = K(a) with K(a) being a geodesic with K(0) = K, [d/dt K](0) = H
    """
    d = K.shape[0]
    rK = K.shape[1]
    pdim = K.shape[2]
    n = pdim * rK
    K = K.reshape(d, n, pdim)
    K_new = K.copy()
    AR_mat = np.zeros((2 * pdim, 2 * pdim)).astype(np.complex128)
    for i in range(d):
        Q, R = np.linalg.qr((np.eye(n) - K[i] @ K[i].T.conj()) @ H[i])
        AR_mat[:pdim, :pdim] = K[i].T.conj() @ H[i]
        AR_mat[pdim:, :pdim] = R
        AR_mat[:pdim, pdim:] = -R.T.conj()
        MN = eigy_expm(-a * AR_mat) @ np.eye(2 * pdim, pdim)
        K_new[i] = K[i] @ MN[:pdim, :] + Q @ MN[pdim:, :]
    return K_new.reshape(d, rK, pdim, pdim)


def lineobjf_isom_geodesic(a, H, K, E, rho, J, y):
    """Compute objective function at position on geodesic

    Parameters
    ----------
    a : float
        Geodesic curve parameter
    H : numpy array
        Element of the tangent space at K and local direction of the geodesic
    K : numpy array
        Current position
    E : numpy array
        Current POVM estimate
    rho : numpy array
        Current initial state estimate
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        The columns contain the outcome probabilities for different povm elements

    Returns
    -------
    f(a): float
        Objective function value at new position along the geodesic
    """
    d = K.shape[0]
    pdim = K.shape[2]
    r = pdim**2
    K_test = update_K_geodesic(K, H, a)
    X_test = np.einsum("ijkl,ijnm -> iknlm", K_test, K_test.conj()).reshape((d, r, r))
    return objf(X_test, E, rho, J, y)


def update_A_geodesic(A, H, a):
    """Compute a new point on the geodesic for the POVM parametrization

    Parameters
    ----------
    A : numpy array
        Current position
    H : numpy array
        Element of the tangent space at A and local direction of the geodesic
    a : float
        Geodesic curve parameter

    Returns
    -------
    A_new: numpy array
        New position given by A_new = A(a) with A(a) being a geodesic with A(0) = A, [d/dt A](0) = H
    """
    n_povm = A.shape[0]
    pdim = A.shape[1]
    n = pdim * n_povm
    A = A.reshape(n, pdim)
    H = H.reshape(n, pdim)
    A_new = A.copy()
    AR_mat = np.zeros((2 * pdim, 2 * pdim)).astype(np.complex128)
    Q, R = np.linalg.qr((np.eye(n) - A @ A.T.conj()) @ H)
    AR_mat[:pdim, :pdim] = A.T.conj() @ H
    AR_mat[pdim:, :pdim] = R
    AR_mat[:pdim, pdim:] = -R.T.conj()
    MN = eigy_expm(-a * AR_mat) @ np.eye(2 * pdim, pdim)
    A_new = A @ MN[:pdim, :] + Q @ MN[pdim:, :]
    return A_new.reshape(n_povm, pdim, pdim)


def update_B_geodesic(B, H, a):
    """Compute a new point on the geodesic for the initial state parametrization

    Parameters
    ----------
    B : numpy array
        Current initial state parametrization
    H : numpy array
        Element of the tangent space at B and local direction of the geodesic
    a : float
        Geodesic curve parameter

    Returns
    -------
    A_new: numpy array
        New position given by B_new = B(a) with B(a) being a geodesic with B(0) = B,
        [d/dt B](t=0) = H
    """
    pdim = B.shape[0]
    n = pdim**2
    B = B.reshape(n)
    H = H.reshape(n)
    B_temp = B.copy()
    AR_mat = np.zeros((2, 2)).astype(np.complex128)
    R = np.linalg.norm((np.eye(n) - np.outer(B, B.T.conj())) @ H)
    Q = ((np.eye(n) - np.outer(B, B.T.conj())) @ H) / R
    AR_mat[0, 0] = B.T.conj() @ H
    AR_mat[1, 0] = R
    AR_mat[0, 1] = -R.T.conj()
    MN = eigy_expm(-a * AR_mat) @ np.array([1, 0])
    B_temp = B * MN[0] + Q * MN[1]
    return B_temp.reshape(pdim, pdim)


def lineobjf_A_geodesic(a, H, X, A, rho, J, y):
    """Compute objective function at position on geodesic for POVM parametrization

    Parameters
    ----------
    a : float
        Geodesic curve parameter
    H : numpy array
        Element of the tangent space at A and local direction of the geodesic
    X : numpy array
        Current gate estimate
    A : numpy array
        Current position
    rho : numpy array
        Current initial state estimate
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        The columns contain the outcome probabilities for different povm elements

    Returns
    -------
    f(a): float
        Objective function value at new position along the geodesic
    """
    n_povm = A.shape[0]
    A_test = update_A_geodesic(A, H, a)
    E_test = np.array([(A_test[i].T.conj() @ A_test[i]).reshape(-1) for i in range(n_povm)])
    return objf(X, E_test, rho, J, y)


def lineobjf_B_geodesic(a, H, X, E, B, J, y):
    """Compute objective function at position on geodesic for the initial state parametrization

    Parameters
    ----------
    a : float
        Geodesic curve parameter
    H : numpy array
        Element of the tangent space at B and local direction of the geodesic
    X : numpy array
        Current gate estimate
    E : numpy array
        Current POVM estimate
    B : numpy array
        Current initial state parametrization
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        The columns contain the outcome probabilities for different povm elements

    Returns
    -------
    f(a): float
        Objective function value at new position along the geodesic
    """
    B_test = update_B_geodesic(B, H, a)
    rho_test = (B_test @ B_test.T.conj()).reshape(-1)
    return objf(X, E, rho_test, J, y)


def lineobjf_A_B(a, v, delta_v, X, C, y, J, argument):
    """Compute objective function at translated position

    Parameters
    ----------
    a : float
        Step size
    v : numpy array
        Current position
    delta_v : numpy array
        Step direction
    X : numpy array
        Current gate estimate
    C : numpy array
        Current initial state/POVM estimate that is not updated
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        The columns contain the outcome probabilities for different povm elements
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    argument: string
        Takes the following options: "E" or "rho", and indicates which gate set
        component is optimized over

    Returns
    -------
    f(a): float
        Objective function value at new position v + a*delta_v

    Notes
    -------
    This function is used for the line search with linear updates v_new = v + a*delta_v,
    where v can be either the POVM estimate or the state estimate.
    """
    v_test = v - a * delta_v
    if argument == "rho":
        rho_test = (v_test @ v_test.T.conj()).reshape(-1)
        return objf(X, C, rho_test, J, y)
    if argument == "E":
        E_test = (v_test @ v_test.T.conj()).reshape(-1)
        return objf(X, E_test, C, J, y)
    raise ValueError("The <argument> variable takes either E or rho")


def Hess_evals(K, E, rho, y, J):
    """Compute eigenvalues of the euclidean Hessian

    Parameters
    ----------
    K : numpy array
        Current position
    E : numpy array
        Current POVM estimate
    rho : numpy array
        Current initial state estimate
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        The columns contain the outcome probabilities for different povm elements
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence

    Returns
    -------
    evals: 1D numpy array
        Eigenvalues of the euclidean Hessian for the Kraus operators at position (K,E,rho)
    """
    d = K.shape[0]
    rK = K.shape[1]
    pdim = K.shape[2]
    r = pdim**2
    n = d * rK * r
    H = np.zeros((2 * n, 2 * n)).astype(np.complex128)
    X = np.einsum("ijkl,ijnm -> iknlm", K, K.conj()).reshape((d, r, r))
    _, dM10, dM11 = dK_dMdM(X, K, E, rho, J, y, d, r, rK)
    dd, dconjd = ddM(X, K, E, rho, J, y, d, r, rK)

    A00 = dM11.reshape(n, n) + np.einsum("ijklmnop->ikmojlnp", dconjd).reshape(n, n)
    A10 = dM10.reshape(n, n) + np.einsum("ijklmnop->ikmojlnp", dd).reshape(n, n)
    A11 = A00.conj()
    A01 = A10.conj()

    H[:n, :n] = A00
    H[:n, n:] = A01
    H[n:, :n] = A10
    H[n:, n:] = A11

    evals, _ = eigh(H)
    return evals



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Compressive gate set tomography

The full benchmark executes the following steps:
    1) Generation of circuits to measure
    2) Running of said circuits on the backend
    3) Conversion of the ideal gate set into numpy arrays to be processed by mGST
    4) Optional generation of an initialization for mGST
    5) Running mGST to get first gate set estimates
    6) Gauge optimization to the target gates using pyGSTi's gauge optimizer
    7) Optional rerun of mGST on bootstrap data to produce error bars
    8) Generation of error measures and result tables
    9) Generation of matrix plots for the reconstructed gate set
"""

from typing import Any, Dict, List, Tuple, Type, Union

import numpy as np
from qiskit import QuantumCircuit
from qiskit.circuit.library import CZGate, RGate
import xarray as xr

from iqm.benchmarks.benchmark import BenchmarkConfigurationBase
from iqm.benchmarks.benchmark_definition import Benchmark, add_counts_to_dataset
from iqm.benchmarks.compressive_gst.gst_analysis import mgst_analysis
from iqm.benchmarks.logging_config import qcvv_logger
from iqm.benchmarks.utils import (
    perform_backend_transpilation,
    retrieve_all_counts,
    set_coupling_map,
    submit_execute,
    timeit,
)
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase
from mGST import additional_fns, qiskit_interface
from mGST.qiskit_interface import add_idle_gates, remove_idle_wires


class CompressiveGST(Benchmark):
    """
    SPAM-robust characterization of a set of quantum gates
    """

    analysis_function = staticmethod(mgst_analysis)

    def __init__(self, backend: IQMBackendBase, configuration: "GSTConfiguration"):
        """Construct the compressive_gst class.

        Args:
            backend (IQMBackendBase): the backend to execute the benchmark on
            configuration (GSTConfiguration): the configuration of the benchmark
        """
        super().__init__(backend, configuration)

        self.qubit_layouts = parse_layouts(configuration.qubit_layouts)
        self.num_qubits = len(self.qubit_layouts[0])
        self.pdim = 2**self.num_qubits
        self.num_povm = self.pdim

        self.gate_set, self.gate_labels, self.num_gates = parse_gate_set(configuration, self.num_qubits)

        if configuration.opt_method not in ["GD", "SFN", "auto"]:
            raise ValueError("Invalid optimization method, valid options are: GD, SFN, auto")
        if configuration.opt_method == "auto":
            if (self.num_qubits == 2 and configuration.rank > 2) or self.num_qubits > 2:
                self.opt_method = "GD"
            else:
                self.opt_method = "SFN"
        else:
            self.opt_method = configuration.opt_method

        if configuration.max_iterations == "auto":
            if self.opt_method == "SFN":
                self.max_iterations = [100, 100]
            if self.opt_method == "GD":
                self.max_iterations = [250, 250]
        elif isinstance(configuration.max_iterations, list):
            self.max_iterations = configuration.max_iterations
        if configuration.batch_size == "auto":
            self.batch_size = 30 * self.pdim
        else:
            self.batch_size = configuration.batch_size

        self.y, self.J = (
            np.empty((self.num_povm, self.configuration.num_circuits)),
            np.empty((self.configuration.num_circuits, self.num_povm)),
        )  # format used by mGST
        self.bootstrap_results = List[Tuple[np.ndarray]]  # List of GST outcomes from bootstrapping

    @staticmethod
    def name() -> str:
        return "compressive_GST"

    @timeit
    def generate_meas_circuits(self) -> None:
        """Generate random circuits from the gate set

        The random circuits are distributed among different depths ranging from L_MIN
        to L_MAX, both are configurable and stored in self.configuration.seq_len_list.
        No transpilation other than mapping to the desired qubits is performed,
        as the gates need to be executed axactly as described for GST to give
        meaningful results

        Returns:
            circuit_gen_transp_time: float
                The time it took to generate and transpile the circuits

        """
        # Calculate number of short and long circuits
        N_short = int(np.ceil(self.configuration.num_circuits / 2))
        N_long = int(np.floor(self.configuration.num_circuits / 2))
        L_MIN, L_CUT, L_MAX = self.configuration.seq_len_list

        # Generate random circuits
        J = additional_fns.random_seq_design(self.num_gates, L_MIN, L_CUT, L_MAX, N_short, N_long)

        # Reduce circuits to exclude negative placeholder values
        gate_circuits = [list(seq[seq >= 0]) for seq in J]

        # Sequences in GST are applied to the state from right to left
        self.J = np.array(J)[:, ::-1]

        unmapped_qubits = list(np.arange(self.num_qubits))
        raw_qc_list = qiskit_interface.get_qiskit_circuits(
            gate_circuits, self.gate_set, self.num_qubits, unmapped_qubits
        )

        for qubits in self.qubit_layouts:
            coupling_map = set_coupling_map(qubits, self.backend, physical_layout="fixed")

            # Perform transpilation to backend
            qcvv_logger.info(
                f"Will transpile all {self.configuration.num_circuits} circuits according to fixed physical layout"
            )
            transpiled_qc_list, _ = perform_backend_transpilation(
                raw_qc_list,
                self.backend,
                qubits,
                coupling_map=coupling_map,
                qiskit_optim_level=0,
                optimize_sqg=False,
                drop_final_rz=False,
            )
            # Saving raw and transpiled circuits in a consistent format with other benchmarks
            self.untranspiled_circuits.update({str(qubits): raw_qc_list})
            self.transpiled_circuits.update({str(qubits): transpiled_qc_list})

    def add_configuration_to_dataset(self, dataset):  # CHECK
        """
        Creates an xarray.Dataset and adds the circuits and configuration metadata to it

        Args:
            self: Source class
        Returns:
            dataset: xarray.Dataset to be used for further data storage
        """
        # Adding configuration entries and class variables, prioritizing the latter in case of conflicts
        for key, value in (self.configuration.__dict__ | self.__dict__).items():
            if key == "benchmark":  # Avoid saving the class objects
                dataset.attrs[key] = value.name()
            elif key == "backend":
                dataset.attrs[key] = value.name
            else:
                dataset.attrs[key] = value
        dataset.attrs["gauge_weights"] = dict({f"G%i" % i: 1 for i in range(self.num_gates)}, **{"spam": 0.1})

    def execute(self, backend) -> xr.Dataset:
        """
        The main GST execution routine
        """

        dataset = xr.Dataset()
        qcvv_logger.info(f"Now generating {self.configuration.num_circuits} random GST circuits...")

        # Generate circuits
        self.generate_meas_circuits()

        # Submit all
        all_jobs: Dict = {}
        for qubit_layout in self.qubit_layouts:
            transpiled_circuit_dict = {tuple(qubit_layout): self.transpiled_circuits[str(qubit_layout)]}
            all_jobs[str(qubit_layout)], _ = submit_execute(
                transpiled_circuit_dict,
                backend,
                self.configuration.shots,
                self.calset_id,
                max_gates_per_batch=self.configuration.max_gates_per_batch,
            )
        # Retrieve all
        qcvv_logger.info(f"Now executing the corresponding circuit batch")
        for qubit_layout in self.qubit_layouts:
            counts, _ = retrieve_all_counts(all_jobs[str(qubit_layout)])
            dataset, _ = add_counts_to_dataset(counts, str(qubit_layout), dataset)

        self.add_configuration_to_dataset(dataset)
        return dataset


class GSTConfiguration(BenchmarkConfigurationBase):
    """Compressive GST configuration base.

    Attributes:
        benchmark (Type[Benchmark]): GHZBenchmark
        qubit_layouts (Sequence[Sequence[int]]): A sequence (e.g., Tuple or List) of sequences of
            physical qubit layouts, as specified by integer labels, where the benchmark is meant to be run.
        gate_set (Union[str, List[Type[QuantumCircuit]]]): The gate set in given either as a list of qiskit quantum
            cirucuits, or as one of the predefined gate sets "1QXYI", "2QXYCZ", "2QXYCZ_extended", "3QXYCZ".
            A gate set should be tomographically complete, meaning from the specified gates and the vacuum state, one
            should be able to prepare states that form a frame for the state space. A practical sufficient condition
            is that the gate set is able the generate all combinations of local Pauli eigenstates.
        num_circuits (int): How many random circuits are generated from the gate set. Guidelines on choosing this value:
            - At least 50 circuits for a single qubit gate set with 3 gates.
            - At least 400 circuits for a two qubit gate set with 6 gates.
            - At least 2000 circuits for a three qubit gate set with 9 gates.
            The number of random circuits needed is expected to grow linearly in the number of gates and exponentially
            in the number of qubits.
        rank (int): The Kraus rank of the reconstruction. Choose rank=1 for coherent error analysis and rank<=dim**2
            generally.
        shots (int): The number of measurement shots per circuit.
            * Default: 1024
        gate_labels (Union[Dict, None]): Names of the gates in the gate set. Used for plots and result tables.
            * Default: None
        seq_len_list (list[int]): Three numbers controling the depth of the random sequences. The first is the minimal
            depth, the last is the maximal depth and the middle number specifies a cutoff depth below which all possible
            sequences are selected.
            * Default: [1, 8, 14]
        from_init (bool): Whether the target gate set is used as an initialization to the mGST algorithm.
            * Default: True
        max_inits (int): If from_init = False, random initial points are tried and this parameter limits the amount of
            retries.
            * Default: 20
        opt_method (str): Which optimization method is used, can be either of "GD" or "SFN", for gradient descent or
            saddle free Newton, respectively.
            * Default: "auto" (Method is automatically selected based on qubit count and rank)
        max_iterations (Union[str, List[int]]): How many iterations to run the optimization algorithm for. Accepted
            values are either "auto" or a list of two integers. The first specifies the number of iterations for the
            optimization on batches or circuit data, while the second specifies the number of iterations on the full
            data for all circuits.
            * Default: "auto" (Numbers are automatically selected based on qubit count and rank)
        convergence_criteria (Union[str, List[float]]): Two parameters which determine when the optimization algorithm
            terminates. The first is a multiplier which specifies how close the cost function should get to a threshold
            for the given shot noise in order for the optimization to be considered "successful". Values in [2,10] are
            usually sensible. The second parameter sets the relative change in cost function between
            two consecutive iterations, below which the algorithm terminates.
            * Default: [4, 1e-4]
        batch_size (Union[str, int]): The number of circuits per batch in the optimization. This hyperparamters is
            automatically set and determines the convergence behaviour. A smaller batch size reduces runtime but can
            lead to erratic jumps in parameter space and lack of convergence.
            * Default: "auto"
        bootstrap_samples (int): The number of times the optimization algorithm is repeated on fake data to estimate
            the uncertainty via bootstrapping.
    """

    benchmark: Type[Benchmark] = CompressiveGST
    qubit_layouts: Union[List[int], List[List[int]]]
    gate_set: Union[str, List[Any]]
    num_circuits: int
    rank: int
    shots: int = 2**10
    gate_labels: Union[List, None] = None
    seq_len_list: list[int] = [1, 8, 14]
    from_init: bool = True
    max_inits: int = 20
    opt_method: str = "auto"
    max_iterations: Union[str, List[int]] = "auto"
    convergence_criteria: Union[str, List[float]] = [4, 1e-4]
    batch_size: Union[str, int] = "auto"
    bootstrap_samples: int = 0


def parse_layouts(qubit_layouts: Union[List[int], List[List[int]]]) -> List[List[int]]:
    """Checks for correct setting of qubit_layouts in the configuration and return a correct type

    Args:
        qubit_layouts: List[List[[int]] or List[int]
            The qubit_layouts on the backend where the gates are defined on

    Returns:
        qubit_layouts: List[List[[int]]
            A properly typed qubit_layout if no Error was raised
    """
    if all(isinstance(qubits, int) for qubits in qubit_layouts):
        return [qubit_layouts]  # type: ignore
    if all(isinstance(qubits, List) for qubits in qubit_layouts):
        if all(len(x) == len(qubit_layouts[0]) for x in qubit_layouts):  # type: ignore
            return qubit_layouts  # type: ignore
        raise ValueError(
            "Qubit layouts are of different size, currently only specifying "
            "qubit layouts with the same number of qubits is supported."
        )
    raise ValueError(
        "The qubit_layouts parameters needs to be one of List[int] for a single layout"
        " or List[List[int]] for multiple layouts."
    )


def parse_gate_set(
    configuration: GSTConfiguration, num_qubits
) -> Tuple[List[QuantumCircuit], Dict[str, Dict[int, str]], int]:
    """
    Handles different gate set inputs and produces a valid gate set

    Args:
        configuration: BenchmarkConfigurationBase
            Configuration class containing variables
        num_qubits: int
            The number of qubits on which the gate set is defined

    Returns:
        gate_set: List[QuantumCircuit]
            A list of gates defined as quantum circuit objects
        gate_labels: List[Dict[int, str]]
            A dictionary with gate names for each layout
        num_gates: int
            The number of gates in the gate set

    """
    if isinstance(configuration.gate_set, str) and configuration.gate_set not in [
        "1QXYI",
        "2QXYCZ",
        "2QXYCZ_extended",
        "3QXYCZ",
    ]:
        raise ValueError(
            "No gate set of the specified name is implemented, please choose among "
            "1QXYI, 2QXYCZ, 2QXYCZ_extended, 3QXYCZ."
        )
    if configuration.gate_set in ["1QXYI", "2QXYCZ", "2QXYCZ_extended", "3QXYCZ"]:
        gate_set, gate_labels, num_gates = create_predefined_gate_set(configuration.gate_set, num_qubits)
        return gate_set, gate_labels, num_gates

    if isinstance(configuration.gate_set, list):
        gate_set = configuration.gate_set
        num_gates = len(gate_set)
        if configuration.gate_labels is None:
            gate_labels = {i: f"Gate %i" % i for i in range(num_gates)}
        else:
            if configuration.gate_labels:
                if len(configuration.gate_labels) != num_gates:
                    raise ValueError(
                        f"The number of gate labels (%i) does not match the number of gates (%i)"
                        % (len(configuration.gate_labels), num_gates)
                    )
                gate_labels = dict(enumerate(configuration.gate_labels))
        return gate_set, gate_labels, num_gates

    raise ValueError(
        f"Invalid gate set, choose among 1QXYI, 2QXYCZ, 2QXYCZ_extended,"
        f" 3QXYCZ or provide a list of Qiskti circuits to define the gates."
    )


def create_predefined_gate_set(gate_set, num_qubits) -> Tuple[List[QuantumCircuit], Dict, int]:
    """Create a list of quantum circuits corresponding to a predefined gate set.

    The circuits are assigned to the specified qubit_layouts on the backend only during transipilation, so the qubit labels
    at this stage may not represent the actual qubit labels on the backend.
    Args:
        gate_set: str
            The name of the gate set
    Returns:
        gates: List[QuantumCircuit]
            The gate set as a list of circuits
        gate_labels_dict: Dict[int, str]
            The names of gates, i.e. "Rx(pi/2)" for a pi/2 rotation around the x-axis.
        num_gates: int
            The number of gates in the gate set

    """
    unmapped_qubits = list(np.arange(num_qubits))

    if gate_set == "1QXYI":
        gate_list = [RGate(1e-10, 0), RGate(0.5 * np.pi, 0), RGate(0.5 * np.pi, np.pi / 2)]
        gates = [QuantumCircuit(num_qubits, 0) for _ in range(len(gate_list))]
        gate_qubits = [[0], [0], [0]]
        for i, gate in enumerate(gate_list):
            gates[i].append(gate, gate_qubits[i])
        gate_labels = ["Idle", "Rx(pi/2)", "Ry(pi/2)"]
    elif gate_set == "2QXYCZ":
        gate_qubits = [[0], [1], [0], [1], [0, 1]]
        gates = [QuantumCircuit(num_qubits, 0) for _ in range(5)]
        gates[0].append(RGate(0.5 * np.pi, 0), [0])
        gates[1].append(RGate(0.5 * np.pi, 0), [1])
        gates[2].append(RGate(0.5 * np.pi, np.pi / 2), [0])
        gates[3].append(RGate(0.5 * np.pi, np.pi / 2), [1])
        gates[4].append(CZGate(), [0, 1])
        gate_labels = ["Rx(pi/2)", "Rx(pi/2)", "Ry(pi/2)", "Ry(pi/2)", "CZ"]
    elif gate_set == "2QXYCZ_extended":
        gate_qubits = [[0], [1], [0], [1], [0, 1], [0, 1], [0, 1], [0, 1], [0, 1]]
        gates = [QuantumCircuit(num_qubits, 0) for _ in range(9)]
        gates[0].append(RGate(0.5 * np.pi, 0), [0])
        gates[1].append(RGate(0.5 * np.pi, 0), [1])
        gates[2].append(RGate(0.5 * np.pi, np.pi / 2), [0])
        gates[3].append(RGate(0.5 * np.pi, np.pi / 2), [1])
        gates[4].append(RGate(0.5 * np.pi, 0), [0])
        gates[4].append(RGate(0.5 * np.pi, 0), [1])
        gates[5].append(RGate(0.5 * np.pi, 0), [0])
        gates[5].append(RGate(0.5 * np.pi, np.pi / 2), [1])
        gates[6].append(RGate(0.5 * np.pi, np.pi / 2), [0])
        gates[6].append(RGate(0.5 * np.pi, 0), [1])
        gates[7].append(RGate(0.5 * np.pi, np.pi / 2), [0])
        gates[7].append(RGate(0.5 * np.pi, np.pi / 2), [1])
        gates[8].append(CZGate(), [[0], [1]])
        gate_labels = [
            "Rx(pi/2)",
            "Rx(pi/2)",
            "Ry(pi/2)",
            "Ry(pi/2)",
            "Rx(pi/2)--Rx(pi/2)",
            "Rx(pi/2)--Ry(pi/2)",
            "Ry(pi/2)--Rx(pi/2)",
            "Ry(pi/2)--Ry(pi/2)",
            "CZ",
        ]
    elif gate_set == "3QXYCZ":
        gate_list = [
            RGate(0.5 * np.pi, 0),
            RGate(0.5 * np.pi, 0),
            RGate(0.5 * np.pi, 0),
            RGate(0.5 * np.pi, np.pi / 2),
            RGate(0.5 * np.pi, np.pi / 2),
            RGate(0.5 * np.pi, np.pi / 2),
            CZGate(),
            CZGate(),
        ]
        gates = [QuantumCircuit(num_qubits, 0) for _ in range(len(gate_list))]
        gate_qubits = [[0], [1], [2], [0], [1], [2], [0, 1], [0, 2]]
        for i, gate in enumerate(gate_list):
            gates[i].append(gate, gate_qubits[i])
        gate_labels = ["Rx(pi/2)", "Rx(pi/2)", "Rx(pi/2)", "Ry(pi/2)", "Ry(pi/2)", "Ry(pi/2)", "CZ", "CZ"]
    else:
        raise ValueError(
            f"Invalid gate set, choose among 1QXYI, 2QXYCZ, 2QXYCZ_extended,"
            f" 3QXYCZ or provide a list of Qiskti circuits to define the gates."
        )

    gates = add_idle_gates(gates, unmapped_qubits, gate_qubits)
    gates = [remove_idle_wires(qc) for qc in gates]

    gate_label_dict = dict(enumerate(gate_labels))

    return gates, gate_label_dict, len(gates)



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Lmfit for multiple datasets, generalized and extended version of the shown example in the lmfit documentation:
https://lmfit.github.io/lmfit-py/examples/example_fit_multi_datasets.html#sphx-glr-examples-example-fit-multi-datasets-py

"""

from inspect import signature

from lmfit import Parameters


def eval_func_single_dataset(func, params, i, x):
    """Returns the evaluation of the fit function for a single dataset"""

    _, fit_param_names = get_param_names_from_func_signature(func)

    # assume parameter object has existing parameter names with a suffix _{i}
    args = [params[f"{name}_{i + 1}"] for name in fit_param_names]

    return func(x, *args)


def get_param_names_from_func_signature(func):
    """Gets the function parameter names from its signature"""
    param_names = list(signature(func).parameters.keys())
    independent_param_name = param_names[0]
    fit_param_names = param_names[1:]
    return independent_param_name, fit_param_names


# use model .post_fit= to calculate fidelities.


def multi_dataset_residual(params, x, data, func):
    """Calculate total residual for fits of func to several data sets."""
    ndata, _ = data.shape
    resid = 0.0 * data[:]

    # make residual per line
    for i in range(ndata):
        resid[i, :] = data[i, :] - eval_func_single_dataset(func, params, i, x)

    # now flatten this to a 1D array; this will be needed by minimize()
    return resid.flatten()


def create_multi_dataset_params(func, data, initial_guesses=None, constraints=None, simultaneously_fit_vars=None):
    """Generates lmfit Parameter object with parameters for each line to fit"""
    fit_params = Parameters()

    _, fit_param_names = get_param_names_from_func_signature(func)

    for i, _ in enumerate(data):
        for name in fit_param_names:
            fit_params.add(f"{name}_{i + 1}")

            # set initial guesses if provided
            if initial_guesses is not None:
                if name in initial_guesses:
                    fit_params[f"{name}_{i + 1}"].value = initial_guesses[name]
            # set constraints if provided
            if constraints is not None:
                if name in constraints:
                    fit_params[f"{name}_{i + 1}"].max = constraints[name]["max"]
                    fit_params[f"{name}_{i + 1}"].min = constraints[name]["min"]

            # force params to be fit as a single parameter across datasets
            if simultaneously_fit_vars is not None:
                if name in simultaneously_fit_vars:
                    if i > 0:
                        fit_params[f"{name}_{i + 1}"].expr = f"{name}_{1}"

    return fit_params
